from ..preproc.quality import high_freq_data_detection, saturated_data_detection
import os
import warnings
import pickle as pkl
import re
import json
import numpy as np
from pandas import DataFrame
import xarray as xr

def parse_file_info(file_path):
    """parse_file_info

    Parses file strings for goose_wireless ECoG and LFP signal data into data parameters.

    Args:
        file_path (str): path to the file's location

    Returns:
        exp_file_name (str): JSON experiment data file path
        mask_file_name (str): binary data mask file path
        microdrive_name (str): string name of the microdrive type used to collect data in file_path
        rec_type (str): recording modality reflected in this file ('ECOG', 'LFP', etc.)
    """

    file_name = os.path.basename(file_path)
    data_file_noext = os.path.splitext(file_name)[0]
    data_file_parts = data_file_noext.split('.')
    if len(data_file_parts) == 3:
        rec_id, microdrive_name, rec_type = data_file_parts
    else:
        rec_id, microdrive_name, _, rec_type = data_file_parts
    data_dir = os.path.dirname(file_path)
    exp_file_name = os.path.join(data_dir,rec_id + ".experiment.json")
    mask_file_name = os.path.join(data_dir,data_file_noext + ".mask.pkl")
    return exp_file_name, mask_file_name, microdrive_name, rec_type

def load_experiment_data(exp_file_name):
    """load_experiment_data

    Reads experiment metadata from an experiment JSON file. Returns the complete data structure as a dictionary and returns electrode data as a pandas DataFrame.

    Args:
        exp_file_name (str): JSON experiment data file path

    Returns:
        experiment (dict): dict data object containing experiment metadata. See lab documentation for more information.
        electrode_df (DataFrame): pandas DataFrame containing microdrive electrode information. Individual channels are indexed along columns, column names are electrode IDs.
    """

    assert os.path.exists(exp_file_name), f'inferred experiment file not found at {exp_file_name}'
    with open(exp_file_name,'r') as f:
        experiment = json.load(f)
    electrode_df = DataFrame(experiment['hardware']['microdrive'][0]['electrodes'])
    electrode_df = DataFrame.join(electrode_df,DataFrame(list(electrode_df.position)))
    del electrode_df['position']
    return experiment, electrode_df

def load_mask_data(mask_file_name):
    """load_mask_data

    Loads binary mask data from recording mask files. Binary True values indicate "bad" or noisy data not used in analyses.

    Args:
        mask_file_name (str): file path to binary mask file

    Returns:
        mask (numpy.array): numpy array of binary values. Length is equal to the number of time points in the respective data array.
    """

    assert os.path.exists(mask_file_name), f'inferred mask file not found at {mask_file_name}'
    with open(mask_file_name,'rb') as f:
        return pkl.load(f)

# def read_lfp(file_path,t_range=(0,-1)):
#     """read_lfp

#     reads data from a structured binary *lfp file in the goose wireless dataset.

#     Args:
#         file_path (str): file path to data file
#         t_range (listlike, optional): Start and stop times to read data. (0, -1) reads the entire file. Defaults to (0,-1).

#     Returns:
#         da (numpy.array): numpy array of multichannel recorded neural activity saved in file_path
#         mask (numpy.array): numpy array of binary mask values
#     """

#     # get local experiment, mask files
#     exp_file_name, mask_file_name, microdrive_name, rec_type = parse_file_info(file_path)

#     # load experiment data
#     experiment, electrode_df = load_experiment_data(exp_file_name)

#     # load mask data
#     mask = load_mask_data(mask_file_name)

#     # get parameters: srate
#     dsmatch = re.search(r'clfp_ds(\d+)',rec_type)
#     if rec_type == 'raw':
#         srate = experiment['hardware']['acquisition']['samplingrate']
#         data_type = np.ushort
#         reshape_order = 'F'
#     elif rec_type == 'lfp':
#         srate = 1000
#         data_type = np.float32
#         reshape_order = 'F'
#     elif rec_type == 'clfp':
#         srate = 1000
#         data_type = np.float32
#         reshape_order = 'F'
#     elif dsmatch:
#         # downsampled data - get srate from name
#         srate = int(dsmatch.group(1))
#         data_type = np.float32
#         reshape_order = 'C' # files created with np.tofile which forces C ordering.

#     # get microdrive parameters
#     microdrive_name_list = [md['name'] for md in experiment['hardware']['microdrive']]
#     microdrive_idx = [md_idx for md_idx, md in enumerate(microdrive_name_list) if microdrive_name == md][0]
#     microdrive_dict = experiment['hardware']['microdrive'][microdrive_idx]
#     num_ch = len(microdrive_dict['electrodes'])

#     # get file size information
#     data_type_size = data_type().nbytes
#     file_size = os.path.getsize(file_path)
#     n_offset_samples = np.round(t_range[0]*srate)
#     n_offset_bytes = n_offset_samples*data_type_size
#     n_all = int(np.floor(file_size/num_ch/data_type_size))
#     n_stop = n_all if t_range[1] == -1 else np.min((np.round(t_range[1]*srate),n_all))
#     n_read = n_stop-n_offset_samples

#     # read signal data
#     data = read_from_file(
#         file_path,
#         data_type,
#         num_ch,
#         n_read,
#         n_offset_bytes,
#         reshape_order=reshape_order
#     )

#     # create xarray from data and channel information
#     da = xr.DataArray(
#         data.T,
#         dime = ('sample','ch'),
#         coords = {
#             'ch': electrode_df.label,
#             'x_pos': ('ch', electrode_df.x),
#             'y_pos': ('ch', electrode_df.y),
#             'row': ('ch', electrode_df.row),
#             'col': ('ch', electrode_df.col),
#         },
#         attrs = {'srate': srate}
#     )

#     return da, mask

# wrapper to read and handle clfp ECOG data
def load_ecog_clfp_data(data_file_name,t_range=(0,-1),exp_file_name=None,mask_file_name=None,compute_mask=True):
    """load_ecog_clfp_data

    Load ECoG data file from a goose wireless dataset file.

    Args:
        data_file_name (str): file path to data file
        t_range (listlike, optional): Start and stop times to read data. (0, -1) reads the entire file. Defaults to (0,-1).
        exp_file_name (str, optional): File path to experiment data JSON file.
        mask_file_name (str, optional): File path to data quality mask file. Defaults to None.
        compute_mask (bool, optional): Compute a data quality mask array if no mask file is given or found. Defaults to True.

    Raises:
        NameError: If experiment file cannot be found, NameError is raised.
        NameError: If mask file cannot be found, NameError is raised.

    Returns:
        data (nt x nch): numpy array of multichannel ECoG data
        mask (numpy.array): binary mask indicating bad data samples
        exp (dict): dictionary of experiment data
    """

    # get file path, set ancillary data file names
    exp_file_name, mask_file_name, microdrive_name, rec_type = parse_file_info(data_file_name)

    # check for experiment file, load if valid, exit if not.
    if os.path.exists(exp_file_name):
        with open(exp_file_name,'r') as f:
            experiment = json.load(f)
    else:
        raise NameError(f'Experiment file {exp_file_name} either invalid or not found. Aborting Process.')

    # get srate
    dsmatch = re.search(r'clfp_ds(\d+)',rec_type)
    if rec_type == 'raw':
        srate = experiment['hardware']['acquisition']['samplingrate']
        data_type = np.ushort
        reshape_order = 'F'
    elif rec_type == 'lfp':
        srate = 1000
        data_type = np.float32
        reshape_order = 'F'
    elif rec_type == 'clfp':
        srate = 1000
        data_type = np.float32
        reshape_order = 'F'
    elif dsmatch:
        # downsampled data - get srate from name
        srate = int(dsmatch.group(1))
        data_type = np.float32
        compute_mask = False
        reshape_order = 'C' # files created with np.tofile which forces C ordering. Sorry!
    else:
        raise NameError(f'File type {rec_type}.dat not recognized. Aborting read process.')

    # get microdrive parameters
    microdrive_name_list = [md['name'] for md in experiment['hardware']['microdrive']]
    microdrive_idx = [md_idx for md_idx, md in enumerate(microdrive_name_list) if microdrive_name == md][0]
    microdrive_dict = experiment['hardware']['microdrive'][microdrive_idx]
    num_ch = len(microdrive_dict['electrodes'])

    exp = {"srate":srate,"num_ch":num_ch}

    data_type_size = data_type().nbytes
    file_size = os.path.getsize(data_file_name)
    n_offset_samples = np.round(t_range[0]*srate)
    n_offset = n_offset_samples*data_type_size
    n_all = int(np.floor(file_size/num_ch/data_type_size))
    if t_range[1] == -1:
        n_stop = n_all
    else:
        n_stop = np.min((np.round(t_range[1]*srate),n_all))
    n_read = n_stop-n_offset_samples

    # load data
    print("Loading data file:")
    # n_offset value is the number of bytes to skip
    # n_read value is the number of items to read (by data type)
    data = read_from_file(data_file_name,data_type,num_ch,n_read,n_offset,
                          reshape_order=reshape_order)
    if rec_type == 'raw': # correct uint16 encoding errors
        data = np.array(data,dtype=np.float32)
        for ch_idx in range(num_ch):
            is_neg = data[ch_idx,:] > 2**15
            data[ch_idx,is_neg] = data[ch_idx,is_neg] - (2**16 - 1)

    # check for mask file, load if valid, compute if not
    if os.path.exists(mask_file_name):
        with open(mask_file_name,"rb") as mask_f:
            mask = pkl.load(mask_f)
    elif compute_mask:
        print(f"No mask data file found for {data_file_name}")
        print("Computing data masks:")
        hf_mask,_ = high_freq_data_detection(data,srate)
        _,sat_mask_all = saturated_data_detection(data,srate)
        sat_mask = np.any(sat_mask_all,axis=0)
        mask = {"hf":hf_mask,"sat":sat_mask}
        # save mask data to current directory
        print(f"Saving mask data for {data_file_name} to {mask_file_name}")
        with open(mask_file_name,"wb") as mask_f:
            pkl.dump(mask,mask_f)
    else:
        mask = []

    return data, exp, mask

# read T seconds of data from the start of the recording:
def read_from_start(data_file_path,data_type,n_ch,n_read):
    """read_from_start

    Read data from goose wireless data file. Reads a fixed number of samples from the start of the recording.

    Args:
        data_file_path (str): file path to data file
        data_type (numeric type): numpy numeric type reflecting the variable encoding in data_file_path
        n_ch (int): number of channels saved in data_file_path
        n_read (int): number of time points to read from data_file_path

    Returns:
        data (np.array): numpy array of neural recording data saved in data_file_path
    """
    data_file = open(data_file_path,"rb")
    data = np.fromfile(data_file,dtype=data_type,count=n_read*n_ch)
    data = np.reshape(data,(n_ch,n_read),order='F')
    data = data.T
    data_file.close()

    return data

# read some time from a given offset
def read_from_file(data_file_path,data_type,n_ch,n_read,n_offset,reshape_order='F'):
    """read_from_file

    Reads recorded neural activity from a goose_wireless file.

    Args:
        data_file_path (str): file path to data file
        data_type (numeric type): numpy numeric type reflecting the variable encoding in data_file_path
        n_ch (int): Number of channels in data_file_path
        n_read (int): Number of data samples read from data_file_path
        n_offset (int): Offset point defining where data reading starts
        reshape_order (str, optional): Data reshaping order. Defaults to 'F'

    Returns:
        data (np.array): numpy array of neural activity stored in data_file_path
    """
    data_file = open(data_file_path,"rb")
    if np.version.version >= "1.17": # "offset" field not added until later installations
        data = np.fromfile(data_file,dtype=data_type,count=n_read*n_ch,
                           offset=n_offset*n_ch)
    else:
        warnings.warn("'offset' feature not available in numpy <= 1.13 - reading from the top",FutureWarning)
        data = np.fromfile(data_file,dtype=data_type,count=n_read*n_ch)
    data = np.reshape(data,(n_ch,n_read),order=reshape_order)
    data_file.close()
    data = data.T

    return data

# read variables from the "experiment.mat" files
def get_exp_var(exp_data,*args):
    """get_exp_var

    Generate a list of variable names from a .MAT formatted experiment data

    Args:
        exp_data (dict): MAT file data dict

    Returns:
        var_names (list): list of variable names in exp_data
    """
    out = exp_data.copy()
    for k, var_name in enumerate(args):
        if k > 1:
            out = out[None][0][None][0][var_name]

        else:
            out = out[var_name]

    return out

# data filtration code




import numpy as np

class MetadataArray(np.ndarray):
    """Numpy array plus metadata."""

    def __new__(cls, input_array, metadata=None):
        obj = np.asarray(input_array).view(cls)
        obj.metadata = metadata
        return obj

    def __array_finalize__(self, obj):
        if obj is None: return
        self.metadata = getattr(obj, 'metadata', None)


# accllr.py
# 
# Latency calculations based on the paper:
#
# Banerjee A, Dean HL, Pesaran B. A likelihood method for computing selection times 
# in spiking and local field potential activity. J Neurophysiol. 2010 Dec;104(6):3705-20. 
# doi: 10.1152/jn.00036.2010. Epub 2010 Sep 8. https://pubmed.ncbi.nlm.nih.gov/20884767/

import numpy as np
from scipy.stats import norm
from statsmodels.stats.multitest import fdrcorrection
from tqdm.auto import tqdm
import multiprocessing as mp
import warnings

from ..utils import first_nonzero

def calc_llr_gauss(lfp, lfp_mean_1, lfp_mean_2, lfp_sigma_1, lfp_sigma_2):
    '''
    Calculate log likelihood ratio of lfp data belonging to one of two 
    gaussian models with no history.
    
    Args:
        lfp ((nt,) array): single trial lfp data
        lfp_mean_1 ((nt,) array): mean across trials of lfp data from condition 1
        lfp_mean_2 ((nt,) array): mean across trials of lfp data from condition 2
        lfp_sigma_1 (float): variance of lfp data from condition 1
        lfp_sigma_2 (float): variance of lfp data from condition 2
    
    Returns:
        (nt,) array: log likelihood ratio 
    
    '''
    res_1 = lfp - lfp_mean_1
    res_2 = lfp - lfp_mean_2

    llr = (np.log(lfp_sigma_2) 
          - np.log(lfp_sigma_1)
          + res_2**2/(2*lfp_sigma_2**2)
          - res_1**2./(2*lfp_sigma_1**2))
    
    return llr
   
def calc_accllr_lfp(lfp_altcond, lfp_nullcond, lfp_altcond_lowpass, lfp_nullcond_lowpass, common_variance=True):
    '''
    Accumulated log likelihood for single channel LFP data.
    
    Note: 
        As quoted from Banerjee et al. (2010), the common_variance parameter should
        be set to True:
        "Since AccLLR changes for LFP activity are scaled by the noise variance for 
        each condition, if the noise variances for each condition differ, the AccLLRs 
        scale differently and this leads to the need for different upper and lower 
        bounds at the discrimination stage. To allow the simplicity of using the same 
        bound for both conditions (see the following subsection), we assume that the 
        variance of activity under each condition is the same."
        
    Note:
        Unsure whether the lowpass filtered version of the lfp should be used in 
        place of the lfp entirely. That is how the Pesaran lab code works, however
        it doesn't seem to make sense to throw away the raw lfp, so here I use
        the lowpass filtered version for the mean estimate across trials but use 
        the raw lfp for single trials.
        
    Note:
        Finally, there is a difference between how the cumulative sum is calculated
        here versus in the Pesaran lab code. Here we include the first value of
        the LLR as the first value of the accLLR. However, the Pesaran lab truncates
        the accLLR so that it starts at the second sum.
    
    Args:
        lfp_altcond ((nt, ntrial) array): lfp data for alternative condition trials
        lfp_nullcond ((nt, ntrial) array): lfp data for null condition trials. The number of null condition
            trials must be the same as the number of alternative condition trials.
        lfp_altcond_lowpass ((nt, ntrial) array): low-pass filtered copy of alternative condition trials. If
            desired, just pass the lfp_altcond again to avoid using a low-pass filtered version for model building.
        lfp_nullcond_lowpass ((nt, ntrial) array): low-pass filtered copy of null condition trials
        common_variance (bool, optional): calculate a shared variance of event and null lfp (see notes). Default True.
    
    Returns:
        tuple: tuple containing:
            | **accllr_altcond ((nt, ntrial) array):** accumulated log-likelihood for alterative condition trials
            | **accllr_nullcond ((nt, ntrial) array):** accumulated log-likelihood for null condition trials
    '''    
    assert lfp_altcond.shape == lfp_nullcond.shape

    nt = lfp_altcond.shape[0]
    n_trial = lfp_altcond.shape[1]
    
    lfp_mean_altcond = np.mean(lfp_altcond_lowpass, axis=1)
    lfp_mean_nullcond = np.mean(lfp_nullcond_lowpass, axis=1)
        
    lfp_sigma_altcond = np.std(lfp_altcond.T - lfp_mean_altcond, ddof=1) # ddof=1 is the behavior of MATLAB
    lfp_sigma_nullcond = np.std(lfp_nullcond.T - lfp_mean_nullcond, ddof=1)
    lfp_sigma_both = (lfp_sigma_altcond + lfp_sigma_nullcond)/2
        
    accllr_altcond = np.zeros((nt, n_trial))
    accllr_nullcond = np.zeros((nt, n_trial))
    
    for idx_trial in range(n_trial):
        
        loo_trials = np.ones((n_trial,), dtype='bool')
        loo_trials[idx_trial] = 0
        
        lfp_mean_altcond_loo = np.mean(lfp_altcond_lowpass[:,loo_trials], axis=1)
        lfp_mean_nullcond_loo = np.mean(lfp_nullcond_lowpass[:,loo_trials], axis=1)
        
        if common_variance:
            llr_altcond = calc_llr_gauss(lfp_altcond[:,idx_trial], 
                                         lfp_mean_altcond_loo, lfp_mean_nullcond, 
                                         lfp_sigma_both, lfp_sigma_both)
            llr_nullcond = calc_llr_gauss(lfp_nullcond[:,idx_trial], lfp_mean_altcond, 
                                          lfp_mean_nullcond_loo, 
                                          lfp_sigma_both, lfp_sigma_both)
        else:
            llr_altcond = calc_llr_gauss(lfp_altcond[:,idx_trial], 
                                         lfp_mean_altcond_loo, lfp_mean_nullcond, 
                                         lfp_sigma_altcond, lfp_sigma_nullcond)
            llr_nullcond = calc_llr_gauss(lfp_nullcond[:,idx_trial], lfp_mean_altcond, 
                                          lfp_mean_nullcond_loo,
                                          lfp_sigma_altcond, lfp_sigma_nullcond)

        accllr_altcond[:, idx_trial] = np.nancumsum(llr_altcond)
        accllr_nullcond[:, idx_trial] = np.nancumsum(llr_nullcond)
            
    return accllr_altcond, accllr_nullcond

# Try using ROC approach from Qiao et al 2020
def detect_accllr(accllr, upper, lower):
    '''
    Calculate the probability of upper, lower, and unknown level detections. This
    version is present only for readability, since it is very slow. See 
    :func:`~aopy.analysis.accllr.detect_accllr_fast`.
    
    Args:
        accllr (nt, ntrial): the accllr timeseries to test across trials
        upper (float): upper level
        lower (float): lower level
        
    Returns:
        tuple: tuple containing:
            | **p (3,):** probability of upper, lower, and unknown level detections
            | **selection_idx (ntrial):** index at which accllr crosses upper threshold 
                (or nan if missed) for each trial
    '''
    ntrial = accllr.shape[1]
    selection_idx = np.zeros((ntrial,))
    n_upper = 0
    n_lower = 0
    n_unknown = 0
    
    for tr_idx in range(ntrial):
        lower_hit = accllr[:,tr_idx] < lower
        upper_hit = accllr[:,tr_idx] > upper
        if np.any(lower_hit) and np.any(upper_hit):
            
            # both thresholds hit, count whichever was first
            if np.where(lower_hit)[0][0] < np.where(upper_hit)[0][0]:
                n_lower += 1
                selection_idx[tr_idx] = np.where(lower_hit)[0][0]
            else:
                n_upper += 1
                selection_idx[tr_idx] = np.where(upper_hit)[0][0]
            
        elif np.any(lower_hit):
            
            # lower threshold hit
            n_lower += 1
            selection_idx[tr_idx] = np.nan
            
        elif np.any(upper_hit):
            
            # upper threshold hit
            n_upper += 1
            selection_idx[tr_idx] = np.where(upper_hit)[0][0]
            
        else:
            
            # no hit
            n_unknown += 1
            selection_idx[tr_idx] = np.nan

    
    p = np.array([
        n_upper/ntrial,
        n_lower/ntrial,
        n_unknown/ntrial
    ])
    
    return p, selection_idx

# Write a faster version of the detection algorithm
def detect_accllr_fast(accllr, upper, lower):
    '''
    Calculate the probability of upper, lower, and unknown level detections. This
    faster algorithm avoids looping over every single trial, greatly speeding up
    computations of accllr. 
    
    Args:
        accllr (nt, ntrial): the accllr timeseries to test across trials
        upper (float): upper level
        lower (float): lower level
        
    Returns:
        tuple: tuple containing:
            | **p (3,):** probability of upper, lower, and unknown level detections
            | **selection_idx (ntrial):** index at which accllr crosses upper threshold 
                (or nan if missed) for each trial
    '''
    nt = accllr.shape[0]
    n_trial = accllr.shape[1]
    
    # Find the first value above the threshold for all trials at once
    lower_hit = accllr < lower
    upper_hit = accllr > upper
    lower_hit_st = first_nonzero(lower_hit, axis=0, all_zeros_val=nt+1) # if no hit, then set nt+1
    upper_hit_st = first_nonzero(upper_hit, axis=0, all_zeros_val=nt+1)

    # Break lower-upper ties
    lower_first = lower_hit_st < upper_hit_st
    st = np.where(lower_first, np.nan, upper_hit_st)
    
    unknown = st == nt+1 # no threshold was hit
    st[st == nt+1] = np.nan

    # Sum the counts to calculate probabilites
    n_lower = np.count_nonzero(lower_first & ~unknown)
    n_unknown = np.count_nonzero(unknown)
    n_upper = n_trial - n_lower - n_unknown
    p = np.array([
        n_upper/n_trial,
        n_lower/n_trial,
        n_unknown/n_trial
    ])
    
    return p, st

def calc_accllr_performance(accllr_altcond, accllr_nullcond, nlevels=200):
    '''
    Calculate the probabilities and selection times of accllr trials for a number of
    different levels, evenly spaced between 0 and the max value of the accllr series. 
    
    Args:
        event_accllr (nt, ntrial): accllr timeseries for alternative condition trials
        null_accllr (nt, ntrial): accllr timeseries for null condition trials
        nlevels (int, optional): number of levels to calculate. Defaults to 200.
        
    Returns:
        tuple: tuple containing:
            | **p_altcond (nlevels,3):** probability of upper, lower, and unknown level detections
                for the alternative condition at each detection level
            | **p_nullcond (nlevels,3):** probability of upper, lower, and unknown level detections
                for the null condition at each detection level
            | **selection_idx (nlevels):** index at which accllr crosses upper threshold 
                (or nan if missed) for each trial, averaged across trials
            | **levels (nlevels):** levels used for calculation of probabilities
    '''
    max_accllr = np.nanmax([accllr_altcond, accllr_nullcond]) # ignore nan
    levels = np.linspace(max_accllr/nlevels,max_accllr,nlevels)

    p_altcond = np.ones((nlevels,3))
    p_nullcond = np.ones((nlevels,3))
    selection_idx = []
    for idx_level, level in enumerate(levels):
        p_altcond[idx_level,:], selection_idx_altcond = detect_accllr_fast(accllr_altcond, level, -level)
        p_nullcond[idx_level,:], _ = detect_accllr_fast(accllr_nullcond, level, -level)
        if np.all(selection_idx_altcond!=selection_idx_altcond):
            selection_idx.append(np.nan)
        else:
            selection_idx.append(np.nanmean(selection_idx_altcond))
        
    return p_altcond, p_nullcond, selection_idx, levels

def choose_best_level(p_altcond, p_nullcond, levels):
    '''
    Given a list of probabilities for upper, lower, and unknown detections at 
    various levels, select the level with the highest difference in correct 
    detection and incorrect detection.

    Args:
        p_altcond ((ntrial, 3) array): [upper, lower, unknown] detections for 
            trials from the alternative condition
        p_nullcond ((ntrial, 3) array): [upper, lower, unknown] detections for 
            trials from the null condition
        levels (nlevels): levels used for calculation of probabilities

    Returns:
        float: the best level to use
    '''
    p_correct_detect = (p_altcond[:,0]+p_nullcond[:,1])/2
    p_incorrect_detect = (p_nullcond[:,0]+p_altcond[:,1])/2
    p_diff = p_correct_detect - p_incorrect_detect
    p_max_idx = np.argmax(p_diff)
    p_max = p_diff[p_max_idx]
    level = levels[p_max_idx]
    return level

def compute_midrank(x):
    '''
    Computes midranks.
    
    Adapted from: https://github.com/yandexdataschool/roc_comparison

    Args:
       x (npt): data to compute midrank over
    
    Returns:
       (npt): array of midranks
    '''
    x_sorting = np.argsort(x)
    x_sorted = x[x_sorting]
    npt = len(x)
    mid_sorted = np.zeros(npt, dtype=float)
    i = 0
    while i < npt:
        j = i
        while j < npt and x_sorted[j] == x_sorted[i]:
            j += 1
        mid_sorted[i:j] = 0.5*(i + j - 1)
        i = j
    midrank = np.empty(npt, dtype=float)
    midrank[x_sorting] = mid_sorted + 1
    return midrank

def calc_delong_roc_variance(ground_truth, predictions):
    '''
    Computes ROC AUC variance for a single set of predictions using the
    fast version of DeLong's method for computing the covariance of
    unadjusted AUC.

    Adapted from https://github.com/Netflix/vmaf/

    Args:
        ground_truth (nt): array of 0 and 1 ground truth classes
        predictions (nt): array of floats of the probability of being class 1
       
    Returns:
        tuple: tuple containing:
            | **auc (float):** area under the curve after ROC analysis
            | **cov (float):** variance of the predicted auc
 
    Reference:
        @article{sun2014fast,
            title={Fast Implementation of DeLong's Algorithm for
                  Comparing the Areas Under Correlated Receiver Oerating
                  Characteristic Curves},
            author={Xu Sun and Weichao Xu},
            journal={IEEE Signal Processing Letters},
            volume={21},
            number={11},
            pages={1389--1393},
            year={2014},
            publisher={IEEE}
        }
    '''
    assert np.array_equal(np.unique(ground_truth), [0, 1])
    order = (-ground_truth).argsort()
    label_1_count = int(ground_truth.sum())
    predictions_sorted_transposed = predictions[np.newaxis, order]

    # Short variables are named as they are in the paper
    m = label_1_count
    n = predictions_sorted_transposed.shape[1] - m
    positive_examples = predictions_sorted_transposed[:, :m]
    negative_examples = predictions_sorted_transposed[:, m:]
    k = predictions_sorted_transposed.shape[0]

    tx = np.empty([k, m], dtype=float)
    ty = np.empty([k, n], dtype=float)
    tz = np.empty([k, m + n], dtype=float)
    for r in range(k):
        tx[r, :] = compute_midrank(positive_examples[r, :])
        ty[r, :] = compute_midrank(negative_examples[r, :])
        tz[r, :] = compute_midrank(predictions_sorted_transposed[r, :])
    aucs = tz[:, :m].sum(axis=1) / m / n - float(m + 1.0) / 2.0 / n
    v01 = (tz[:, :m] - tx[:, :]) / n
    v10 = 1.0 - (tz[:, m:] - ty[:, :]) / m
    sx = np.cov(v01)
    sy = np.cov(v10)
    delongcov = sx / m + sy / n
    assert len(aucs) == 1, "There is a bug in the code, please forward this to the developers"
    return aucs[0], delongcov

def calc_accllr_roc(accllr_altcond, accllr_nullcond):
    '''
    ROC analysis on accllr data. Groups the data according to class (alt or null) and computes
    area under the curve (AUC) using the DeLong method.
    
    Args:
        accllr_altcond ((ntrial,) array): a single accumulated llr across trials for alternative
            condition trials. Usually we pick the end of the accumulation window as the timepoint 
            of interest.
        accllr_nullcond((ntrial,) array): a single accumulated llr across trials for null
            condition trials.

    Returns:
        tuple: tuple containing:
            | **auc (float):** area under the curve
            | **se (float):** error across trials of the auc
    '''
    group = np.concatenate([np.ones((len(accllr_altcond),)), np.zeros((len(accllr_nullcond),))])
    data = np.concatenate([accllr_altcond, accllr_nullcond])

    auc, var = calc_delong_roc_variance(group, data)
    return auc, np.sqrt(var)

def calc_accllr_st_single_ch(data_altcond_ch, data_nullcond_ch, lowpass_altcond_ch, lowpass_nullcond_ch,
                       modality, bin_width, nlevels, verbose_out=False):
    '''
    Calculate accllr selection time for a single channel

    Args:
        data_altcond_ch ((nt, ntrial)): lfp data for trials from the alternative condition
        data_nullcond_ch ((nt, ntrial)): lfp data for trials from the null condition
        lowpass_altcond_ch ((nt, ntrial) array): low-pass filtered copy of alternative condition trials. If
            desired, just pass the lfp_altcond again to avoid using a low-pass filtered version for model building.
        lowpass_nullcond_ch ((nt, ntrial) array): low-pass filtered copy of null condition trials
        modality (str): type of data being inputted ("lfp", "spikes", etc.)
        bin_width (float): bin width of input activity, or 1./samplerate for lfp data
        nlevels (int): number of levels at which to test accllr performance
        verbose_out (bool, optional): Include probabilities of detection and accllr in the output. Defaults to False.

    Returns:
        tuple: tuple containing:
            | **selection_time (nch, ntrials):** time (in s) at which each trial 
            | **roc_auc (nch,):** area under the curve from receiver operating characteristic analysis
            | **roc_se (nch,):** error across trials of the auc analysis
    '''
    # Calculate accllr for null and altcond
    if modality == 'lfp':
        accllr_altcond, accllr_nullcond = calc_accllr_lfp(data_altcond_ch, data_nullcond_ch, 
                                                          lowpass_altcond_ch, lowpass_nullcond_ch)
    else:
        raise ValueError("AccLLR currently only supports LFP")
        
    # Calculate appropriate levels to select accLLR decisions
    p_altcond, p_nullcond, _, levels = calc_accllr_performance(accllr_altcond, accllr_nullcond, nlevels)
    level = choose_best_level(p_altcond, p_nullcond, levels)

    # Calculate p values and selection times using the level with highest significance
    p_altcond, selection_idx_altcond = detect_accllr_fast(accllr_altcond, level,-level)
    selection_time_altcond = (selection_idx_altcond)*bin_width

    # Calculate ROC at the end of AccLLR accmulation time
    roc_auc, roc_se = calc_accllr_roc(accllr_altcond[-1,:], accllr_nullcond[-1,:])

    if verbose_out:
        p_nullcond, selection_idx_nullcond = detect_accllr_fast(-accllr_nullcond, level, -level)
        return (accllr_altcond, accllr_nullcond, p_altcond, p_nullcond, 
                selection_time_altcond, roc_auc, roc_se)

    return selection_time_altcond, roc_auc, roc_se

def _calc_accllr_st_worker(data_altcond, data_nullcond, lowpass_altcond, lowpass_nullcond,
                           modality, bin_width, nlevels, parallel):
    '''
    Worker function for calc_accllr_st(). 

    Args:
        data_altcond ((nt, nch, ntrial)): lfp channel data for trials from the alternative condition
        data_nullcond ((nt, nch, ntrial)): lfp channel data for trials from the null condition
        lowpass_altcond ((nt, nch, ntrial) array): low-pass filtered copy of alternative condition trials. If
            desired, just pass the lfp_altcond again to avoid using a low-pass filtered version for model building.
        lowpass_nullcond ((nt, nch, ntrial) array): low-pass filtered copy of null condition trials
        modality (str): type of data being inputted ("lfp", "spikes", etc.)
        bin_width (float): bin width of input activity, or 1./samplerate for lfp data
        nlevels (int): number of levels at which to test accllr performance. 
        parallel (bool): if True, run the computations across channels in parallel.

    Returns:
        tuple: tuple containing:
            | **selection_time (nch, ntrials):** time (in s) at which each trial 
            | **roc_auc (nch,):** area under the curve from receiver operating characteristic analysis
            | **roc_se (nch,):** error across trials of the auc analysis
    '''
    nt = data_altcond.shape[0]
    nch = data_altcond.shape[1]
    ntrials = data_altcond.shape[2]
    
    # Run accllr on the test datasets    
    if parallel:
        pool = mp.Pool(min(mp.cpu_count(), nch))
        
        # call apply_async() without callback
        result_objects = [pool.apply_async(calc_accllr_st_single_ch, 
                          args=(data_altcond[:,ich,:], data_nullcond[:,ich,:], 
                                lowpass_altcond[:,ich,:], lowpass_nullcond[:,ich,:],
                                modality, bin_width, nlevels)) 
                          for ich in range(nch)]

        # result_objects is a list of pool.ApplyResult objects
        results = [r.get() for r in result_objects]
        selection_time_altcond, roc_auc, roc_se = zip(*results)
        selection_time_altcond = np.array(selection_time_altcond)
        roc_auc = np.squeeze(roc_auc)
        roc_se = np.squeeze(roc_se)

        pool.close()
    else:
        selection_time_altcond = np.zeros((nch, ntrials))*np.nan
        roc_auc = np.zeros((nch,))*np.nan
        roc_se = np.zeros((nch,))*np.nan
        for ich in tqdm(range(nch), desc="AccLLR wrapper", leave=False):
            (selection_time_altcond[ich,:], roc_auc[ich], 
             roc_se[ich]) = calc_accllr_st_single_ch(data_altcond[:,ich,:], data_nullcond[:,ich,:], 
                                                     lowpass_altcond[:,ich,:], lowpass_nullcond[:,ich,:],
                                                     modality, bin_width, nlevels)
                
    return selection_time_altcond, roc_auc, roc_se
            
def calc_accllr_st(data_altcond, data_nullcond, lowpass_altcond, lowpass_nullcond, 
                   modality, bin_width, nlevels=None, match_selectivity=False, 
                   match_ch=None, noise_sd_step=1, parallel=True):
    '''
    Calculate accllr selection time for a multiple channels with optional selectivity
    matching. Selectivity is defined by the area under the ROC curve. To match 
    selectivity, we add gaussian noise to channel until selectivity is constant 
    across all channels.

    Based on the paper:

    Banerjee A, Dean HL, Pesaran B. A likelihood method for computing selection times 
    in spiking and local field potential activity. J Neurophysiol. 2010 Dec;104(6):3705-20. 
    doi: 10.1152/jn.00036.2010. Epub 2010 Sep 8. https://pubmed.ncbi.nlm.nih.gov/20884767/

    Note: 
        No noise is added to the models built on the lowpass versions of data. It is unclear how
        this step was performed in the original paper.

    Examples:

        Below are powers in 50-250hz band of three channels of laser-evoked response in motor cortex, 
        aligned to laser onset across 350 trials, 50 ms before and after. We first apply AccLLR without
        selectivity matching:

        .. code-block:: python

            st, roc_auc, roc_se, roc_p_fdrc = accllr.calc_accllr_st(altcond, nullcond, altcond, 
                                                                    nullcond, 'lfp', 1./samplerate)
            accllr_mean = np.nanmean(st, axis=1)

        .. image:: _images/accllr_test_data.png

        The dotted lines are the estimated selection times returned by accllr. Note that the bigger
        responses have faster selection times -- to test whether the estimates are biased by the 
        larger peaks being easier to identify, we can add noise until they match the selectivity of
        the smaller peak. Selectivity is defined by ROC analysis -- our ability to discriminate 
        between laser event trials and null trials.
    
        .. image:: _images/accllr_test_data_match_selectivity.png

        Note that the selection times for the two larger peaks have shifted slightly to the right,
        but are still earlier than the smaller peak, despite having the same (or lower) selectivity.
        Thus, we can be confident that the bigger peaks appear faster than the smaller one.

    Args:
        data_altcond ((nt, nch, ntrial)): lfp channel data for trials from the alternative condition
        data_nullcond ((nt, nch, ntrial)): lfp channel data for trials from the null condition
        lowpass_altcond ((nt, nch, ntrial) array): low-pass filtered copy of alternative condition trials. If
            desired, just pass the lfp_altcond again to avoid using a low-pass filtered version for model building.
        lowpass_nullcond ((nt, nch, ntrial) array): low-pass filtered copy of null condition trials
        modality (str): type of data being inputted ("lfp", "spikes", etc.)
        bin_width (float): bin width of input activity, or 1./samplerate for lfp data
        nlevels (int): number of levels at which to test accllr performance
        match_selectivity (bool, optional): if True, add noise to input data to match selectivity across channels.
            Default False.
        match_ch ((nch,) bool array or None, optional): if set, limit selectivity matching to only these 
            channels. Default None.
        noise_sd_step (float, optional): standard deviation step size to take when adding noise to ch data. 
            Default 1.
        parallel (bool, optional): if True, run the computations across channels in parallel. Default True.

    Returns:
        tuple: tuple containing:
            | **selection_time (nch, ntrials):** time (in s) at which each trial 
            | **roc_auc (nch,):** area under the curve from receiver operating characteristic analysis
            | **roc_se (nch,):** error across trials of the auc analysis
            | **roc_p_fdrc (nch,):** p-value for each channel after false-discovery-rate correction
    '''
    # Calculate an initial analysis
    nt = data_altcond.shape[0]
    nch = data_altcond.shape[1]
    ntrials = data_altcond.shape[2]
    if nlevels is None:
        nlevels = nt
    (selection_time_altcond, 
     roc_auc, roc_se) = _calc_accllr_st_worker(data_altcond, data_nullcond, 
                                               lowpass_altcond, lowpass_nullcond, modality, 
                                               bin_width, nlevels, parallel)
    
    # Optionally perform selectivity matching
    if match_ch is None:
        match_ch = np.ones((nch,), dtype='bool')
    if match_selectivity and not np.any(match_ch):
        warnings.warn("No channels selected. Not attempting selectivity matching.")
        match_selectivity = False
    if match_selectivity:

        # Find selective channel with the lowest ROC AUC
        match_ch_prob = np.min(roc_auc[match_ch])
        print(f"Matching selectivity to {match_ch_prob:0.4f}. Largest selectivity is {np.max(roc_auc[match_ch]):0.4f}")
        match_ch_idx = roc_auc <= match_ch_prob

        # Use the ROC AUC to match selectivity across channels   
        pbar = tqdm(total=nch-np.sum(match_ch_idx), desc="Matching selectivity")
        noise_sd = noise_sd_step
        while np.sum(match_ch_idx) < nch:
            
            # Add noise to non-matched channels
            unmatch_chs_idx = np.arange(nch) # unmatched channel idx
            unmatch_chs_idx = unmatch_chs_idx[~match_ch_idx]

            noise = np.random.normal(0,noise_sd, size=(nt,len(unmatch_chs_idx), ntrials))
            data_altcond[:,unmatch_chs_idx,:] += noise
            data_nullcond[:,unmatch_chs_idx,:] += noise

            # Re-calculate roc 
            (selection_time_altcond[unmatch_chs_idx,:], roc_auc[unmatch_chs_idx], 
            roc_se[unmatch_chs_idx]) = \
                _calc_accllr_st_worker(data_altcond[:,unmatch_chs_idx,:], data_nullcond[:,unmatch_chs_idx,:],
                                       lowpass_altcond[:,unmatch_chs_idx,:], lowpass_nullcond[:,unmatch_chs_idx,:],
                                       modality, bin_width, nlevels, parallel)

            match_ch_idx = roc_auc <= match_ch_prob
            noise_sd += noise_sd_step
            pbar.update(np.sum(match_ch_idx[unmatch_chs_idx]))
            
        pbar.close()

    # Finally calculate an FDR-corrected p-value across all channels
    z = (roc_auc-0.5)/roc_se # null hypothesis auc=0.5 (chance)
    p_uncorrected = norm.sf(z) # one-sided test of auc>0.5
    if nch > 1:
        rej, roc_p_fdrc = fdrcorrection(p_uncorrected, alpha=0.05)
    else:
        roc_p_fdrc = np.array((np.nan,))
    
    return selection_time_altcond, roc_auc, roc_se, roc_p_fdrc


def prepare_erp(erp, erp_lowpass, samplerate, time_before, time_after, 
                window_nullcond, window_altcond):
    '''
    Prepare data for accllr. Given event-related potentials, organize alternative
    and null condition data and subtract the mean baseline from the null condition.

    Args:
        erp ((nt, nch, ntr) array): trial-aligned data
        erp_lowpass ((nt, nch, ntr) array): trial-aligned data lowpass filtered
        samplerate (float): sampling rate of the erps
        time_before (float): time before event in the erp (in seconds)
        time_after (float): time after event in the erp (in seconds)
        window_nullcond ((2,) tuple of float): desired (start, end) of nullcond (in seconds)
        window_altcond ((2,) tuple of float): desired (start, end) of altcond (in seconds)

    Returns:
        tuple: tuple containing:
            | **data_altcond ((nt_before_new, nch, ntr) array):** alternative condition data
            | **data_nullcond ((nt_before_new, nch, ntr) array):** null condition data
            | **lowpass_altcond ((nt_before_new, nch, ntr) array):** alternative condition low-passed data
            | **lowpass_nullcond ((nt_before_new, nch, ntr) array):** null condition low-passed data

    Example:

        .. code-block:: python

            npts = 100
            nch = 50
            ntrials = 30
            align_idx = 50
            onset_idx = 40
            samplerate = 100
            time_before = align_idx/samplerate
            time_after = time_before
            window_nullcond = (-0.4, -0.1)
            window_altcond = (-0.1, 0.3)
            data = np.ones(npts)*10
            data[onset_idx:] = 10 + np.arange(npts-onset_idx)
            data = np.repeat(np.tile(data, (nch,1)).T[:,:,None], ntrials, axis=2)

            data_altcond, data_nullcond, lowpass_altcond, lowpass_nullcond = accllr.prepare_erp(
                data, data, samplerate, time_before, time_after, window_nullcond, window_altcond,
            )

        .. image:: _images/prepare_erp_for_accllr.png


    '''
    assert len(window_nullcond) == 2 and window_nullcond[1] > window_nullcond[0]
    assert len(window_altcond) == 2 and window_altcond[1] > window_altcond[0]
    assert window_nullcond[0] >= -time_before
    assert window_altcond[1] <= time_after
    
    # Find start and end indices
    altcond_start = int((time_before+window_altcond[0])*samplerate)-1
    altcond_dur = window_altcond[1] - window_altcond[0]
    altcond_end = altcond_start + int(altcond_dur*samplerate)
    nullcond_start = int((time_before+window_nullcond[0])*samplerate)
    nullcond_dur = window_nullcond[1] - window_nullcond[0]
    nullcond_end = nullcond_start+int(nullcond_dur*samplerate)
    
    # Extract data
    data_altcond = erp[altcond_start:altcond_end,:,:].copy()
    data_nullcond = erp[nullcond_start:nullcond_end,:,:].copy()
    lowpass_altcond = erp_lowpass[altcond_start:altcond_end,:,:].copy()
    lowpass_nullcond = erp_lowpass[nullcond_start:nullcond_end,:,:].copy()
    
    # Make each trial zero-mean for both stim and baseline
    baseline = np.mean(data_nullcond, axis=0)
    data_altcond -= baseline
    data_nullcond -= baseline
    lowpass_baseline = np.mean(lowpass_nullcond, axis=0)
    lowpass_altcond -= lowpass_baseline
    lowpass_nullcond -= lowpass_baseline

    return data_altcond, data_nullcond, lowpass_altcond, lowpass_nullcond



# setup.py
# setup script for aopy

import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

install_requires = [
    'numpy<2.0',
    'xarray',
    'pandas',
    'psutil',
    'h5py',
    'tables',
    'scikit-learn',
    'statsmodels',
    'nitime',
    'xlrd',
    'openpyxl',
    'matplotlib>=3.9',
    'scipy',
    'PyWavelets',
    'Pillow',
    'seaborn',
    'pyyaml',
    'tqdm',
    'open-ephys-python-tools',
    'aolab-bmi3d>=1.0',
    'sympy'
]

setuptools.setup(
    name="aolab-aopy",
    version="0.7.3",
    author="aoLab",
    author_email="aorsborn@uw.edu",
    description="python code repository for aoLab @UW",
    long_description=long_description,
    url="https://github.com/aolabNeuro/analyze",
    packages=setuptools.find_namespace_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    install_requires=install_requires,
)



import copy
import numpy as np
import numpy.linalg as npla
import scipy.signal as sps
import matplotlib.pyplot as plt
import traceback
from tqdm.auto import tqdm

from .. import precondition
from .. import analysis
from ..utils import print_progress_bar
from ..visualization.base import plot_image_by_time
from ..data.base import load_preproc_exp_data

def detect_bad_exp_data(preproc_dir, subjects, ids, dates):
    '''
    Identifies preprocessed experiment data files that contain errors. Detected errors include: 
        preprocessed data could not be loaded, preprocessed data is missing event timestamps (likely 
        the result of missing ecube data), and preprocessed data contains events that do not match 
        bmi3d_events (likely the result of inaccurately sent digital events). The detected files 
        should be re-preprocessed or otherwise debugged before being used in analysis!

    Args:
        preproc_dir (str): base directory where the files live
        subjects (list of str): Subject name for each recording
        ids (list of int): Block number of Task entry object for each recording
        dates (list of str): Date for each recording

    Returns:
        list of lists: entries (subject, date, id) of each preprocessed experiment data file that was identified to have errors
    '''
    bad_entries = []
    entries = list(zip(subjects, dates, ids))

    for subject, date, te in tqdm(entries): 
        # Load data from bmi3d hdf 
        try:
            exp_data, exp_metadata = load_preproc_exp_data(preproc_dir, subject, te, date)
        except:
            print(f"Entry {subject} {date} {te} could not be loaded.")
            traceback.print_exc()
            bad_entries.append([subject,date,te])
            continue
        
        # Check events and times
        try:
            event_times = exp_data['events']['timestamp']
        except:
            print(f"Entry {subject} {date} {te} is missing event timestamps (likely missing ecube data).")
            print('source files:', exp_metadata['source_files'])
            bad_entries.append([subject,date,te])
            continue
        
        # Check that events are accurate
        if not np.array_equal(exp_data['events']['code'], exp_data['bmi3d_events']['code']):
            print(f"Entry {subject} {date} {te} was excluded due to mismatched sync and bmi3d events (this will likely cause problems).")
            bad_entries.append([subject,date,te])
    
    return bad_entries       

# python implementation of badChannelDetection.m - see which channels are too noisy
def bad_channel_detection(data, srate, num_th=3., lf_c=100., sg_win_t=8., sg_over_t=4., sg_bw = 0.5):
    """bad_channel_detection

    Checks input [nt, nch] data array channel quality

    Args:
        data (nt, nch): numpy array of data
        srate (int): sample rate
        num_th (float, optional): Constant to adjust threshold. Defaults to 3.
        lf_c (int, optional): low frequency cutoff. Defaults to 100.
        sg_win_t (numeric, optional): spectrogram window length. Defaults to 8.
        sg_over_t (numeric, optional): spectrogram overlap length. Defaults to 4.
        sg_bw (float, optional): spectrogram time-half-bandwidth product. Defaults to 0.5.

    Returns:
        bad_ch (nch): logical array indicating bad channels
    """
    
    sg_step_t = sg_win_t - sg_over_t
    assert sg_step_t > 0, 'window length must be greater than window overlap'
    print("Running bad channel assessment:")

    # compute low-freq PSD estimate
    n, p, k = precondition.convert_taper_parameters(sg_win_t, sg_bw)
    fxx, txx, Sxx_low = analysis.calc_mt_tfr(data, n, p, k, srate, step=sg_step_t, fk=lf_c)
    Sxx_low_psd = np.mean(Sxx_low,axis=1)

    psd_var = np.var(Sxx_low_psd,axis=0)
    norm_psd_var = psd_var/npla.norm(psd_var)
    low_var_θ = np.mean(norm_psd_var)/num_th
    bad_ch = norm_psd_var >= low_var_θ

    return bad_ch


def detect_bad_ch_outliers(data, nbins=10000, thr=0.05, numsd=5.0, debug=False, verbose=True):
    '''
    Detect badchannels. This code is originally Amy's code in pesaran lab.
    This function has 2 steps. In the 1st step, it detects tentative bad channels whose SD is less than th or more than 1-th in CDF
    In the 2nd step, it extracts bad channels from tentative bad channels by finding conditions where (SD - median of SD) is outside numsd*SD[~badch]
    
    Args:
        data (nt, nch): neural data
        nbins (int): number of bins to make CDF
        thr (float, optional): threshold in the CDF to detect bad channels for 1st screening. 0.05 means data outisde 5% or 95% in the CDF is regarded as bad channels.
        numsd (float, optional): number of standard deviations above zero to detect bad channels for 2nd screening
        debug (bool, optional): if True, display a figure showing the threshold crossings
        verbose (bool, optional): if True, print bad channels
        
    Returns:
        bad_ch (nch) : logical array indicating bad channels after 2nd screening

    Example:
        .. code-block:: python

            test_data = np.random.normal(10,0.5,(10000, 200))
            test_data[0, 10] = 25
            test_data[5, 150] = 30
            bad_ch = quality.detect_bad_ch_outliers(test_data, nbins=10000, thr=0.05, numsd=5.0, debug=True, verbose=False)

        .. image:: _images/detect_bad_trials.png
    '''
    
    assert thr > 0 and thr < 1, "Threshold must be between 0 and 1"
    assert numsd > 0, "numsd must be more than 0"
    
    sd = np.std(data, axis=0)
    nch = data.shape[1]
    med_sd = np.median(sd)
    
    # 1st screening
    hist, bins = np.histogram(sd, nbins)
    CDF = np.cumsum(hist)/np.sum(hist)
    bottom = bins[np.where(CDF<thr)[0][-1]]
    top = bins[np.where(CDF>1-thr)[0][0]]
    bad_ch1 = (sd < bottom) | (sd > top)
    
    # 2nd screening
    thr2 = numsd*np.std(sd[~bad_ch1], ddof=1)
    sd_from_med = np.abs(sd - med_sd)
    inrange = sd_from_med <= thr2
    bad_ch = bad_ch1 & ~inrange
    
    if debug:
        fig,ax = plt.subplots(1,2,figsize=(11,4),tight_layout=True)
        ax[0].plot(bins[:-1],CDF)
        ax[0].scatter(bins[np.where(CDF<thr)[0]],CDF[CDF<thr],c='g',s=1,label='bad_ch (1st screening)')
        ax[0].scatter(bins[np.where(CDF>1-thr)[0]],CDF[CDF>1-thr],c='g',s=1)
        ax[0].plot([top,top],[0,1],'r--', label='threshold (1st screening)')
        ax[0].plot([bottom,bottom],[0,1],'r--')
        ax[0].set(xlabel='SD',ylabel='CDF',title='1st screening')
        ax[0].legend()
        ax[1].plot(sd_from_med,'.')
        ax[1].plot([0,nch],[thr2,thr2],'r--', label='threshold (2nd screening)')
        ax[1].plot(np.where(bad_ch1)[0], sd_from_med[bad_ch1], 'g.',label='bad_ch (1st screening)')
        ax[1].plot(np.where(bad_ch)[0], sd_from_med[bad_ch], 'r*', label='bad_ch (2nd screening)')
        ax[1].set(ylabel='abs(SD-median of SD)',xlabel='# channels',title='2nd screening')
        ax[1].legend()
    
    if verbose:
        print(f'Bad channels : {np.where(bad_ch)[0]}')
        print(f'The number of bad channels : {np.sum(bad_ch)}')
        
    return bad_ch

def detect_bad_trials(erp, sd_thr=5, ch_frac=0.5, debug=False):
    '''
    Finds trials where a given fraction of channels contain outlier data.
    
    Args:
        erp (nt, nch, ntr): trial-aligned continuous data
        sd_thr (float, optional): number of standard deviations away from the mean to threshold
            bad data. Default 5
        ch_frac (float, optional): fraction (between 0. and 1.) of channels containing bad data to
            consider a trial as bad. Default 0.5
        debug (bool, optional): if True, display a figure showing the threshold crossings
        
    Returns:
        (ntr,) boolean mask: True for bad trials, False for good trials

    Example:
        .. code-block:: python

            nt = 50
            nch = 10
            ntr = 100
            erp = np.random.normal(size=(nt, nch, ntr)) 
            erp[:,:,0] += 10 # entire trial is noisy across all electrodes
            erp[:,:8,1] -= 10 # entire trial is noisy on most electrodes
            erp[0,:,2] += 10 # single timepoint within the trial is noisy on all electrodes
            for t in range(nt):
                erp[t,t%nch,3] -= 10 # single timepoint is noisy but different timepoint for each channel
                    
            bad_trials = detect_bad_trials(erp, sd_thr=5, ch_frac=0.5, debug=True)

        .. image:: _images/detect_bad_trials.png
    '''
    assert erp.ndim == 3
    nt, nch, ntr = erp.shape
    
    median = np.nanmedian(erp, axis=(0,2), keepdims=True)
    sd = np.nanstd(erp, axis=(0,2), keepdims=True)
    
    bad_timepoints = abs(erp - median) > sd_thr*sd
        
    bad_ch_trials = np.any(bad_timepoints, axis=0) # (nch, ntr) mask where any timepoint is bad
    bad_trials = np.sum(bad_ch_trials, axis=0) > ch_frac * nch # trials where most channels have an outlier

    if debug:

        # Highlight bad timepoints across trials
        plt.figure(figsize=(11,4), layout='compressed')
        plt.subplot(1,2,1)
        erp = abs(erp - median)/sd
        erp = np.nanmax(erp, axis=0)
        erp[bad_ch_trials] = np.nan
        trials = np.arange(ntr)
        cmap = copy.copy(plt.get_cmap('viridis'))
        cmap.set_bad(color='w') # set the 'bad' color to white
        im = plot_image_by_time(trials, erp.T, ylabel='channel', cmap=cmap)
        cbar = plt.colorbar(im)
        cbar.set_label('sd')
        plt.xlabel('trial')        
        plt.title('sd over threshold shown in white')
        
        # Plot number of bad channels for each trial
        plt.subplot(1,2,2)
        ch = np.sum(bad_ch_trials, axis=0)
        trial = np.arange(ntr)
        plt.scatter(trial[~bad_trials], ch[~bad_trials], marker='.', color='k', label='good trials')
        plt.scatter(trial[bad_trials], ch[bad_trials], marker='x', color='r', label='bad trials')
        plt.xlabel('trial')
        plt.ylabel('# channels')
        plt.hlines(ch_frac*nch, 0, ntr, linestyles='dashed', color='r')
        plt.title('fraction of channels above threshold')
        plt.legend()
        
    return bad_trials

def detect_bad_timepoints(data, sd_thr=5, ch_frac=0.5, debug=False):
    '''
    Finds timepoints where a given fraction of channels contain outlier data. For best results,
    you may need to first compute the power of the data before using this function.
    
    Args:
        data (nt, nch): continuous data
        sd_thr (float, optional): number of standard deviations away from the mean to threshold
            bad data. Default 5
        ch_frac (float, optional): fraction (between 0. and 1.) of channels containing bad data to
            consider a trial as bad. Default 0.5
        debug (bool, optional): if True, display a figure showing the threshold crossings
        
    Returns:
        (nt,) boolean mask: True for bad timepoints, False for good timepoints

    Example:
        .. code-block:: python

            nt = 200
            nch = 10
            np.random.seed(0)
            data = np.random.normal(size=(nt, nch)) 
            data[0:50,:] += 10 # timepoint is noisy across all electrodes
            data[50:100,8:] -= 10 # timepoint is noisy on most electrodes
            for t in range(100,nt):
                data[t,t%nch] -= 10 # single timepoint is noisy but different timepoint for each channel
                    
            bad_timepoints = quality.detect_bad_timepoints(data, sd_thr=5, ch_frac=0.5, debug=True)

        .. image:: _images/detect_bad_timepoints.png
    '''
    assert data.ndim == 2
    nt, nch = data.shape
    
    median = np.nanmedian(data, axis=0, keepdims=True)
    sd = np.nanstd(data, axis=0, keepdims=True)
    
    bad_timepoints = abs(data - median) > sd_thr*sd
    bad_ch_timepoints = np.sum(bad_timepoints, axis=1) > ch_frac * nch # trials where most channels have an outlier

    if debug:

        # Highlight bad timepoints across trials
        plt.figure(figsize=(11,4), layout='compressed')
        plt.subplot(1,2,1)
        data = abs(data - median)/sd
        data[bad_timepoints] = np.nan
        time = np.arange(nt)
        cmap = copy.copy(plt.get_cmap('viridis'))
        cmap.set_bad(color='w') # set the 'bad' color to white
        im = plot_image_by_time(time, data, ylabel='channel', cmap=cmap)
        cbar = plt.colorbar(im)
        cbar.set_label('sd')
        plt.xlabel('timepoint')
        plt.title('sd over threshold shown in white')
        
        # Plot number of bad channels for each trial
        plt.subplot(1,2,2)
        ch = np.sum(bad_timepoints, axis=1)
        time = np.arange(nt)
        time[~bad_ch_timepoints]
        ch[~bad_ch_timepoints]
        plt.scatter(time[~bad_ch_timepoints], ch[~bad_ch_timepoints], marker='.', color='k', label='good timepoints')
        plt.scatter(time[bad_ch_timepoints], ch[bad_ch_timepoints], marker='x', color='r', label='bad timepoints')
        plt.xlabel('timepoint')
        plt.ylabel('# channels')
        plt.hlines(ch_frac*nch, 0, nt, linestyles='dashed', color='r')
        plt.title('fraction of channels above threshold')
        plt.legend()
        
    return bad_ch_timepoints

# python implementation of highFreqTimeDetection.m - looks for spectral signatures of junk data
def high_freq_data_detection(data, srate, bad_channels=None, lf_c=100., sg_win_t=8., sg_over_t=4., sg_bw=0.5):
    """high_freq_data_detection

    Checks multichannel numpy array data for excess high frequency power. Returns a logical array of time locations in which any channel has excess high power (indicates noise)

    Args:
        data (nt, nch): timerseries data across channels
        srate (numeric): data sampling rate
        bad_channels (boolean array, optional): Array-like of boolean values indicating bad channels. Defaults to None.
        lf_c (numeric, optional): low frequency cutoff. Defaults to 100.

    Returns:
        bad_data_mask (nt): boolean array indicating timepoints with detected high-frequency noise on any channel
        bad_data_mask_all_ch (nt, nch): boolean array indicating time points at which any channel had high-frequency noise
    """

    print("Running high frequency noise detection: lfc @ {0}".format(lf_c))
    [num_samp, num_ch] = np.shape(data)
    bad_data_mask_all_ch = np.zeros((num_samp, num_ch))
    data_t = np.arange(num_samp)/srate
    if not bad_channels:
        bad_channels = np.zeros(num_ch)

    # estimate hf influence, channel-wise
    for ch_i in np.arange(num_ch)[np.logical_not(bad_channels)]:
        print_progress_bar(ch_i,num_ch)
        sg_step_t = sg_win_t - sg_over_t
        assert sg_step_t > 0, 'window length must be greater than window overlap'
        n, p, k = precondition.convert_taper_parameters(sg_win_t, sg_bw)
        fxx, txx, Sxx = analysis.calc_mt_tfr(data[:, ch_i], n, p, k, srate, step=sg_step_t)
        num_freq, = np.shape(fxx)
        num_t, = np.shape(txx)
        Sxx_mean = np.mean(Sxx, axis=1) # average across all windows, i.e. numch x num_f periodogram

        # get low-freq, high-freq data
        low_f_mask = fxx < lf_c # Hz
        high_f_mask = np.logical_not(low_f_mask)
        low_f_mean = np.mean(Sxx_mean[low_f_mask],axis=0)
        low_f_std = np.std(Sxx_mean[low_f_mask],axis=0)
        high_f_mean = np.mean(Sxx_mean[high_f_mask],axis=0)
        high_f_std = np.std(Sxx_mean[high_f_mask],axis=0)

        # set thresholds for high, low freq. data
        low_θ = low_f_mean - 3*low_f_std
        high_θ = high_f_mean + 3*high_f_std

        for t_i, t_center in enumerate(txx):
            low_f_mean_ = np.mean(Sxx[low_f_mask,t_i])
            high_f_mean_ = np.mean(Sxx[high_f_mask,t_i])
            if low_f_mean_ < low_θ or high_f_mean_ > high_θ:
                # get indeces for the given sgram window and set them to "bad:True"
                t_bad_mask = np.logical_and(data_t > t_center - sg_win_t/2, data_t < t_center + sg_win_t/2)
                bad_data_mask_all_ch[t_bad_mask, ch_i] = True

#     bad_ch_θ = 0
#     bad_data_mask = np.sum(bad_data_mask_all_ch,axis=0) > bad_ch_θ
    bad_data_mask = np.any(bad_data_mask_all_ch,axis=1)

    return bad_data_mask, bad_data_mask_all_ch


# py version of noiseByHistogram.m - get upper and lower signal value bounds from a histogram
def histogram_defined_noise_levels(data, nbin=20):
    """histogram_defined_noise_levels

    Automatically determine bandwidth in a signal

    Args:
        data (np.array): single-channel data array
        nbin (int, optional): number of histogram bins. Defaults to 20.

    Returns:
        noise_bounds (tuple): lower, upper bound values
    """
    # remove data in outer bins of the histogram calculation
    hist, bin_edge = np.histogram(data,bins=nbin)
    low_edge, high_edge = bin_edge[1], bin_edge[-2]
    no_edge_mask = np.all([(data > low_edge), (data < high_edge)],axis = 0)
    data_no_edge = data[no_edge_mask]
    # compute gaussian 99% CI estimate from trimmed data
    data_mean = np.mean(data)
    data_std = np.std(data)
    data_CI_lower, data_CI_higher = data_mean - 3*data_std, data_mean + 3*data_std
    # return min/max values from whole dataset or the edge values, whichever is lower
    noise_lower = low_edge if low_edge < data_CI_lower else min(data)
    noise_upper = high_edge if high_edge > data_CI_higher else max(data)

    return (noise_lower, noise_upper)


# py version of saturatedTimeDetection.m - get indeces of saturated data segments
def saturated_data_detection(data, srate, bad_channels=None, adapt_tol=1e-8 ,
                            win_n=20):

    """saturated_data_detection

    Detects saturated data segments in input data array

    Args:
        data (nt, nch): numpy array of multichannel data
        srate (numeric): data sampling rate
        bad_channels (bool array, optional): boolean array indicating bad data channels. Default: None
        adapt_tol (float, optional): detection tolerance. Default: 1e-8
        win_n (int, optional): sample length of detection window. Default: 20

    Returns:
        sat_data_mask (nt): boolean array indicating saturated data detection
        bad_all_ch_mask (nt, nch): boolean array indicated separate channel saturation detected

    """
    print("Running saturated data segment detection:")
    num_samp, num_ch = np.shape(data)
    if not bad_channels:
        bad_channels = np.zeros(num_ch)
    bad_all_ch_mask = np.zeros((num_samp, num_ch))
    data_rect = np.abs(np.float32(data))
    mask = [bool(not x) for x in bad_channels]

    for ch_i in np.arange(num_ch)[mask]:
        print_progress_bar(ch_i, num_ch)
        ch_data = data_rect[:, ch_i]
        θ1 = 50 # initialize threshold value
        θ0 = 0
        h, valc = np.histogram(ch_data, int(np.max(ch_data)))
        val = (valc[1:] + valc[:-1])/2 # computes the midpoints of each bin, valc are the edges
        val = np.floor(val)
        prob_val = h/np.shape(h)[0]

        # estimate midpoint between bimodal distribution for a theshold value
        while np.abs(θ1 - θ0) > adapt_tol:
            θ0 = θ1
            sub_θ_val_mask = val <= θ1
            sup_θ_val_mask = val > θ1
            sub_θ_val_mean = np.sum(np.multiply(val[sub_θ_val_mask], prob_val[sub_θ_val_mask]))/np.sum(prob_val[sub_θ_val_mask])
            sup_θ_val_mean = np.sum(np.multiply(val[np.logical_not(sup_θ_val_mask)], prob_val[np.logical_not(sup_θ_val_mask)]))/np.sum(prob_val[sup_θ_val_mask])
            θ1 = (sub_θ_val_mean + sup_θ_val_mean)/2

        # filter signal, boxcar window
        b_filt = np.ones(win_n)/win_n
        a_filt = 1
        ch_data_filt = sps.lfilter(b_filt,a_filt,ch_data)
        ch_data_filt_sup_θ_mask = ch_data_filt > θ1

        # get histogram-derived noise limits
        n_low, n_high = histogram_defined_noise_levels(ch_data)
        ch_data_low_mask = ch_data < n_low
        ch_data_high_mask = ch_data > n_high
        ch_data_filt_low_mask = np.logical_and(ch_data_filt_sup_θ_mask, ch_data_low_mask)
        ch_data_filt_high_mask = np.logical_and(ch_data_filt_sup_θ_mask, ch_data_high_mask)
        bad_all_ch_mask[:, ch_i] = np.logical_or(ch_data_filt_low_mask, ch_data_filt_high_mask)

        # clear out straggler values
        # I will hold off on implementing this until
#         out_of_range_samp_mask = np.logical_or(ch_data < n_low, ch_data > n_high)

#         for samp_i in np.arange(samp_i)[np.logical_and(out_of_range_samp_mask,np.logical_not(bad_all_ch_mask[i,:]))]:
#             if np.abs(ch_data[samp_i]) >= θ1 and
#             if samp_i < num_samp - srate*45:

#             else:

    num_bad = np.sum(bad_all_ch_mask,axis=1)
    sat_data_mask = num_bad > num_ch/2

    return sat_data_mask, bad_all_ch_mask


# behavior.py
#
# Behavioral metrics code, e.g. trajectory path lengths, eye movement analysis, success rate, etc.

import numpy as np
from scipy import signal

from .base import calc_rolling_average
from .. import preproc
from .. import postproc
from ..data import load_bmi3d_task_codes

'''
Behavioral metrics 
'''
def calc_success_percent(events, start_events=[b"TARGET_ON"], end_events=[b"REWARD", b"TRIAL_END"], success_events=b"REWARD", window_size=None):
    '''
    A wrapper around get_trial_segments which counts the number of trials with a reward event 
    and divides by the total number of trials. This function can either calculated the success percent
    across all trials in the input events, or compute a rolling success percent based on the 'window_size' 
    input argument. 
    
    See also:
        :func:`~aopy.analysis.calc_success_percent_trials`

    Args:
        events (nevents): events vector, can be codes, event names, anything to match
        start_events (int, str, or list, optional): set of start events to match
        end_events (int, str, or list, optional): set of end events to match
        success_events (int, str, or list, optional): which events make a trial a successful trial
        window_size (int, optional): [in number of trials] For computing rolling success perecent. How many trials 
            to include in each window. If None, this functions calculates the success percent across all trials.

    Returns:
        float or array (nwindow): success percent = number of successful trials out of all trials attempted.
    '''
    segments, _ = preproc.get_trial_segments(events, np.arange(len(events)), start_events, end_events)
    trial_success = [np.any(np.isin(success_events, trial)) for trial in segments]

    return calc_success_percent_trials(trial_success, window_size)

def calc_success_percent_trials(trial_success, window_size=None):
    '''
    A wrapper around get_trial_segments which counts the number of trials with a reward event 
    and divides by the total number of trials. This function can either calculated the success percent
    across all trials in the input events, or compute a rolling success percent based on the 'window_size' 
    input argument. 
    
    See also:
        :func:`~aopy.analysis.calc_success_percent`

    Args:
        trial_success ((ntrial,) bool array): boolean array of trials where success is non-zero and failure is zero
        window_size (int, optional): [in number of trials] For computing rolling success perecent. How many trials 
            to include in each window. If None, this functions calculates the success percent across all trials.

    Returns:
        float or array (nwindow): success percent = number of successful trials out of all trials attempted.
    '''
    n_trials = len(trial_success)

    # If requested, calculate success percent across entire input events
    if window_size is None:
        n_success = np.count_nonzero(trial_success)  
        success_percent = n_success / n_trials

    # Otherwise, compute rolling success percent
    else:
        success_percent = calc_rolling_average(trial_success, window_size)

    return success_percent

def calc_success_rate(events, event_times, start_events, end_events, success_events, window_size=None):
    '''
    Calculate the number of successful trials per second with a given trial start and end definition. 
    Inputs are raw event codes and times.

    See also:
        :func:`~aopy.analysis.calc_success_rate_trials`

    Args:
        events (nevents): events vector, can be codes, event names, anything to match
        event_times (nevents): time of events in 'events'
        start_events (int, str, or list, optional): set of start events to match
        end_events (int, str, or list, optional): set of end events to match
        success_events (int, str, or list, optional): which events make a trial a successful trial
        window_size (int, optional): [ntrials] For computing rolling success perecent. How many trials 
            to include in each window. If None, this functions calculates the success percent across all trials.

    Returns:
        float or array (nwindow): success rate [success/s] = number of successful trials completed per second of time between the start event(s) and end event(s).
    '''
    # Get event time information
    segments, times = preproc.get_trial_segments(events, event_times, start_events, end_events)
    trial_acq_time = times[:,1]-times[:,0]
    trial_success = [np.any(np.isin(success_events, trial)) for trial in segments]

    return calc_success_rate_trials(trial_success, trial_acq_time, window_size=window_size)

def calc_success_rate_trials(trial_success, trial_time, window_size=None):
    '''
    Calculate the number of successful trials per second with a given trial start and end definition. 

    See also:
        :func:`~aopy.analysis.calc_success_rate`

    Args:
        trial_success ((ntrial,) bool array): boolean array of trials where success is non-zero and failure is zero
        trial_time ((ntrial,) array): float array of the time taken in each trial (e.g. acquisition time)
        window_size (int, optional): [ntrials] For computing rolling success perecent. How many trials 
            to include in each window. If None, this functions calculates the success percent across all trials.

    Returns:
        float or array (nwindow): success rate [success/s] = number of successful trials completed per second of 
            time between the start event(s) and end event(s).
    '''
    assert len(trial_time) == len(trial_success), "Mismatched trial lengths"

    # Get % of successful trials per window 
    success_perc = calc_success_percent_trials(trial_success, window_size=window_size)
    ntrials = len(trial_time)

    # Determine rolling target acquisition time info 
    if window_size is None:
        nsuccess = success_perc*ntrials
        acq_time = np.sum(trial_time)

    else:
        nsuccess = success_perc
        acq_time = calc_rolling_average(trial_time, window_size)
    
    success_rate = nsuccess / acq_time

    return success_rate

def compute_path_length_per_trajectory(trajectory):
    '''
    This function calculates the path length by computing the distance from all points for a single trajectory. The input trajectry could be cursor or eye trajectory from a single trial. It returns a single value for path length.

    Args:
        trajectory (nt x 2): single trial trajectory, could be a cursor trajectory or eye trajectory

    Returns:
        path_length (float): length of the trajectory
    '''
    lengths = np.sqrt(np.sum(np.diff(trajectory, axis=0)**2, axis=1)) # compute the distance from all points in trajectory
    path_length = np.sum(lengths)
    return path_length

def compute_movement_error(trajectory, target_position, rotation_vector=np.array([1, 0]), error_axis=1):
    """
    Computes movement error of a trajectory relative to the straight line between the origin and a target position. 

    Args:
        trajectory (nt, 2): The trajectory coordinates for each point in time. 
        target_position (2,): Target position coordinates.
        rotation_vector ((2,) array, optional): The vector onto which movement will be
            projected to calculate error. Defaults to np.array([1, 0]).
        error_axis (int, optional): axis (after rotation) along which to compute the error 
            statistics. Default 1.

    Returns:
        (nt,): The error of the trajectory relative to the target position
    """
    assert np.count_nonzero(target_position) > 0, "Please check target position. Must be non-zero"

    rotated_traj = postproc.rotate_spatial_data(trajectory, rotation_vector, target_position)
    return np.array(rotated_traj)[:, error_axis]

def compute_movement_stats(trajectory, target_position, rotation_vector=np.array([1, 0]), error_axis=1, 
                           return_all_stats=False):
    """
    Computes movement statistics of a trajectory relative to a target position. 

    Args:
        trajectory (nt, 2): The trajectory coordinates for each point in time. 
        target_position (2,): Target position coordinates.
        rotation_vector ((2,) array, optional): The vector onto which movement will be
            projected to calculate error. Defaults to np.array([1, 0]).
        error_axis (int, optional): axis (after rotation) along which to compute the error 
            statistics. Default 1.

    Returns:
        tuple: A tuple containing:
            | **mean (float):** The mean error of the trajectory relative to the target position.
            | **std (float):** The variance of the error of the trajectory relative to the target position.
            | **auc (float):** The area under the curve ofthe trajectory relative to the target position.
        additionally, with return_all_stats=True: 
            | **abs_mean (float):** The mean of the absolute value of the trajectory error.
            | **abs_min (float):** The minimum absolute trajectory error.
            | **abs_max (float):** The maximum absolute trajectory error.
            | **abs_auc (float):** The area under the curve of the absolute value of the trajectory error.
            | **sign (float):** 1 if the maximum positive value is bigger than the maximum negative value. -1 otherwise.
            | **signed_min (float):** The minimum value if the sign is 1, otherwise the maximum value of the trajectory error.
            | **signed_max (float):** The maximum value if the sign is 1, otherwise the minimum value of the trajectory error.
            | **signed_abs_mean (float):** The sign multiplied by the absolute value of the mean trajectory error.
        
    """
    dist_ts = compute_movement_error(trajectory, target_position, rotation_vector, error_axis)
    
    # Statistics of the error axis
    mean = np.mean(dist_ts)
    std = np.std(dist_ts)
    auc = np.sum(dist_ts)

    if not return_all_stats:
        return (mean, std, auc)
    
    # Unsigned statistics
    abs_mean = np.mean(np.abs(dist_ts))
    abs_min = np.min(np.abs(dist_ts))
    abs_max = np.max(np.abs(dist_ts))
    abs_auc = np.sum(np.abs(dist_ts))

    # Signed statistics
    sign = -1 if abs(np.min(dist_ts)) > abs(np.max(dist_ts)) else 1 # bigger negative or positive?
    signed_min = np.min(dist_ts) if sign == 1 else np.max(dist_ts)
    signed_max = np.max(dist_ts) if sign == 1 else np.min(dist_ts)
    signed_abs_mean = sign * abs_mean
    
    return (mean, std, auc, abs_mean, abs_min, abs_max, abs_auc, sign, signed_min, signed_max, signed_abs_mean)

def time_to_target(event_codes, event_times, target_codes=list(range(81, 89)) , go_cue_code=32 , reward_code=48):
    '''
    This function calculates reach time to target only on rewarded trials given trial aligned event codes and event times See: :func:`aopy.preproc.base.get_trial_segments_and_times` .

    Note:
        Trials are filtered to only include rewarded trials so that all trials have the same length.

    Args:
        event_codes (list) : trial aligned event codes
        event_times (list) : trial aligned event times corresponding to the event codes. These event codes and event times could be the output of preproc.base.get_trial_segments_and_times()
        target_codes (list) : list of event codes for cursor entering peripheral target 
        go_cue_code (int) : event code for go cue 
        reward_code (int) : event code for reward

    Returns:
      tuple: tuple containing:
        | **reachtime_pertarget (list)**: duration of each segment after filtering
        | **trial_id (list):** target index on each segment
    '''
    tr_T = np.array([event_times[iTr] for iTr in range(len(event_times)) if reward_code in event_codes[iTr]])
    tr_E = np.array([event_codes[iTr] for iTr in range(len(event_times)) if reward_code in event_codes[iTr]])
    leave_center_idx = np.argwhere(tr_E == go_cue_code)[0, 1]
    reach_target_idx = np.argwhere(np.isin(tr_E[0], target_codes))[0][0] # using just the first trial to get reach_target_idx
    reachtime = tr_T[:, reach_target_idx] - tr_T[:, leave_center_idx]
    target_dir = tr_E[:,reach_target_idx]

    return reachtime, target_dir

def calc_segment_duration(events, event_times, start_events, end_events, target_codes=list(range(81, 89)), trial_filter=lambda x:x):
    '''
    Calculates the duration of trial segments. Event codes and event times for this function are raw and not trial aligned.

    Args:
        events (nevents): events vector, can be codes, event names, anything to match
        event_times (nevents): time of events in 'events'
        start_events (int, str, or list, optional): set of start events to match
        end_events (int, str, or list, optional): set of end events to match
        target_codes (list, optional): list of target codes to use for finding targets within trials
        trial_filter (function, optional): function to apply to each trial's events to determine whether or not to keep it

    Returns:
        tuple: tuple containing:
            | **segment_duration (list)**: duration of each segment after filtering
            | **target_codes (list):** target index on each segment
    '''
    trial_events, trial_times = preproc.get_trial_segments(events, event_times, start_events, end_events)
    trial_events, trial_times = zip(*[(e, t) for e, t in zip(trial_events, trial_times) if trial_filter(e)])

    segment_duration = np.array([t[1] - t[0] for t in trial_times])
    target_idx = [np.argwhere(np.isin(te, target_codes))[0][0] for te in trial_events]
    target_codes = np.array([trial_events[trial_idx][idx] for trial_idx, idx in enumerate(target_idx)]) - np.min(target_codes)

    return segment_duration, target_codes

def get_movement_onset(cursor_traj, fs, trial_start, target_onset, gocue, numsd=3.0, butter_order=4, low_cut=20, thr=None):
    '''
    Compute movement onset when cursor speed crosses threshold based on mean and standard deviation in baseline period.
    Speed is estimated from cursor trajectories and low-pass filtered to remove noise.
    Baseline is defined as the period between target onset and gocue because speed still exists soon after the cursor enters the center target.
    
    Args:
        cursor_traj ((ntr,) np object array) : cursor trajectory that begins with the time when the cursor enters the center target
        fs (float) : sampling rate in Hz
        trial_start (ntr) : trial start time (the time when the cursor enters the center target) relative to experiment start time in sec
        target_onset (ntr) : target onset relative to experiment start time in sec
        gocue (ntr) : gocue (the time when the center target disappears) relative to experiment start time in sec
        numsd (float) : for determining threshold at each trial
        butter_order (int) : the order for the butterworth filter
        low_cut (float) : cut off frequency for low pass filter in Hz
        thr (float) : thr when you want to use constant threshold across trials. If thr=None, thr is computed by mean + numsd*std in the period from target onset to gocue.
        
    Returns:
        movement_onset (ntr) : movement onset relative to trial start time (the time when the cursor enters the center target) in sec
    '''
    
    target_from_start = target_onset - trial_start # target onset relative to trial start time
    gocue_from_start = gocue - trial_start # gocue relative to trial start time
    dt = 1/fs
    
    b, a = signal.butter(butter_order, low_cut, btype='lowpass', fs=fs)

    movement_onset = []
    for itr in range(cursor_traj.shape[0]):
        # compute speed
        dist = np.linalg.norm(cursor_traj[itr],axis=1)
        speed_tmp = np.diff(dist)/(1/fs)
        speed_tmp = np.insert(speed_tmp,0,speed_tmp[0]) # complement the first data point
        speed = signal.filtfilt(b, a, speed_tmp, axis=0)
        
        # compute threshold based on mean and std in baseline
        t_cursor = np.arange(dist.shape[0])*dt
        if thr is None:
            baseline_idx = (t_cursor<gocue_from_start[itr]) & (t_cursor>target_from_start[itr])
            baseline_speed = np.mean(speed[baseline_idx])
            baseline_std = np.std(speed[baseline_idx],ddof=1)
            thr = baseline_speed + numsd*baseline_std
        
        # get movement onset
        movement_onset.append(t_cursor[np.where((speed>thr)&(t_cursor>target_from_start[itr]))[0][0]])
        
    return np.array(movement_onset)

def get_cursor_leave_time(cursor_traj, samplerate, target_radius):
    '''
    Compute the times when the cursor leaves the center target radius
    
    Args:
        cursor_traj ((ntr,) np object array) : cursor trajectory that begins with the time when the cursor enters the center target
        fs (float) : sampling rate in Hz
        target_radius (float) : the radius of the center target
        
    Returns:
        cursor_leave_time (ntr): cursor leave times relative to the time when the cursor enters the center target
    '''
    
    ntr = len(cursor_traj)
    cursor_leave_time = []
    
    for itr in range(ntr):
        t_axis = np.arange(cursor_traj[itr].shape[0])/samplerate
        
        dist = np.linalg.norm(cursor_traj[itr],axis=1)
        leave_idx = np.where(dist>target_radius)[0][0]
        cursor_leave_time.append(t_axis[leave_idx])
    
    return np.array(cursor_leave_time)

'''
Continuous tracking behavioral metrics
'''
def calc_tracking_error(user_traj, target_traj):
    '''
    Computes the mean-squared error between the user position and target position over time.

    Args:
        user_traj (nt,ndim): user trajectory over a trial segment
        target_traj (nt,ndim): target trajectory over a trial segment

    Returns:
        float array (ndim,): tracking error in each dimension
    '''
    assert len(user_traj) == len(target_traj), "User and target trajectories must be the same length!"
    return np.mean((user_traj - target_traj)**2, axis=0) # compute mean over time axis

def calc_tracking_in_time(event_codes, event_times, proportion=False):
    '''
    Computes the total amount of time that the cursor is inside the target over a trial segment.

    Args:
        event_codes (nevents,): list of event codes
        event_times (nevents,): list of event times
        proportion (bool, optional): whether to return the time as a proportion of the total trial segment time. 
            Default is False.

    Returns:
        float: amount of time (in seconds) that the cursor was in the target for
    '''   
    # get all the individual times when cursor was inside target
    task_codes = load_bmi3d_task_codes()
    start_events = [task_codes['CURSOR_ENTER_TARGET']]
    end_events = [task_codes['CURSOR_LEAVE_TARGET'], task_codes['REWARD']] # once cursor is in target, cursor can leave or trial can finish  
    cursor_in_target_segment, cursor_in_target_times = preproc.get_trial_segments_and_times(event_codes, event_times, start_events, end_events)

    # add up the individual times
    tracking_in_time = sum([t[1] - t[0] for t in cursor_in_target_times]) # end time of segment - start time of segment

    # optionally return the time as a proportion of the total segment length
    if proportion:
        tracking_in_time = tracking_in_time/(event_times[-1] - event_times[0])
        
    return tracking_in_time


import os
import h5py
from importlib.metadata import version
import datetime

from .base import *
from .bmi3d import parse_bmi3d
from .oculomatic import parse_oculomatic
from .optitrack import parse_optitrack
from .. import postproc
from .. import precondition
from .. import data as aodata
from ..precondition import eye
from ..data import load_ecube_data_chunked, load_ecube_metadata, proc_ecube_data, save_hdf, load_hdf_group, get_hdf_dictionary, get_preprocessed_filename
from ..data import load_preproc_lfp_data, load_preproc_broadband_data, load_preproc_eye_data

'''
proc_* wrappers
'''
def proc_single(data_dir, files, preproc_dir, subject, te_id, date, preproc_jobs, overwrite=False, **kwargs):
    '''
    Preprocess a single recording, given a list of raw data files, into a series of hdf records with the same prefix.
    Args:
        data_dir (str): File directory of collected session data
        files (dict): dict of file names to process in data_dir
        preproc_dir (str): Target directory for processed data
        subject (str): Subject name
        te_id (int): Block number of Task entry object 
        date (str): Date of recording
        preproc_jobs (list): list of proc_types to generate
        overwrite (bool, optional): Overwrite files in result_dir. Defaults to False.
    '''
    if os.path.basename(os.path.normpath(preproc_dir)) != subject:
        preproc_dir = os.path.join(preproc_dir, subject)
    if not os.path.exists(preproc_dir):
        os.mkdir(preproc_dir)
    preproc_dir_base = os.path.dirname(preproc_dir)

    if 'exp' in preproc_jobs:
        print('processing experiment data...')
        exp_filename = get_preprocessed_filename(subject, te_id, date, 'exp')
        proc_exp(
            data_dir,
            files,
            preproc_dir,
            exp_filename,
            overwrite=overwrite,
        )
    if 'eye' in preproc_jobs:
        print('processing eyetracking data...')
        exp_filename = get_preprocessed_filename(subject, te_id, date, 'exp')
        eye_filename = get_preprocessed_filename(subject, te_id, date, 'eye')
        proc_eyetracking(
            data_dir,
            files,
            preproc_dir,
            exp_filename,
            eye_filename,
            overwrite=overwrite,
        )
        eye_data, eye_metadata = load_preproc_eye_data(preproc_dir_base, subject, te_id, date)
        assert 'raw_data' in eye_data.keys(), "No eye data found"
        assert eye_data['raw_data'].shape == (eye_metadata['n_samples'], eye_metadata['n_channels'])
    if 'broadband' in preproc_jobs:
        print('processing broadband data...')
        broadband_filename = get_preprocessed_filename(subject, te_id, date, 'broadband')
        proc_broadband(
            data_dir,
            files,
            preproc_dir,
            broadband_filename,
            overwrite=overwrite
        )
        broadband_data, broadband_metadata = load_preproc_broadband_data(preproc_dir_base, subject, te_id, date)
        assert broadband_data.shape == (broadband_metadata['n_samples'], broadband_metadata['n_channels'])
        files['broadband'] = broadband_filename # for proc_lfp()
    if 'lfp' in preproc_jobs:
        print('processing local field potential data...')
        lfp_filename = get_preprocessed_filename(subject, te_id, date, 'lfp')
        proc_lfp(
            data_dir,
            files,
            preproc_dir,
            lfp_filename,
            overwrite=overwrite,
            filter_kwargs=kwargs # pass any remaining kwargs to the filtering function
        )
        lfp_data, lfp_metadata = load_preproc_lfp_data(preproc_dir_base, subject, te_id, date)
        assert lfp_data.shape == (lfp_metadata['n_samples'], lfp_metadata['n_channels'])

def proc_exp(data_dir, files, result_dir, result_filename, overwrite=False, save_res=True):
    '''
    Process experiment data files. Loads 'hdf' and 'ecube' (if present) data, parses, and 
    prepares experiment data and metadata.

    Note:
        Currently supports BMI3D only. 
    
    The above data is prepared into structured arrays:
        exp_data:
            task ([('cursor', '<f8', (3,)), ('trial', 'u8', (1,)), ('time', 'u8', (1,)), ...])
            state ([('msg', 'S', (1,)), ('time', 'u8', (1,))])
            clock ([('timestamp', 'f8', (1,)), ('time', 'u8', (1,))])
            events ([('timestamp', 'f8', (1,)), ('time', 'u8', (1,)), ('event', 'S32', (1,)), 
                ('data', 'u2', (1,)), ('code', 'u2', (1,))])
            trials ([('trial', 'u8'), ('index', 'u8'), (condition_name, 'f8', (3,)))])
        exp_metadata:
            source_dir (str)
            source_files (dict)
            bmi3d_start_time (float)
            n_cycles (int)
            n_trials (int)
            <other metadata from bmi3d>
    
    Args:
        data_dir (str): where the data files are located
        files (dict): dictionary of filenames indexed by system
        result_filename (str): where to store the processed result
        overwrite (bool): whether to remove existing processed files if they exist

    Returns:
        None
    '''   
    # Check if a processed file already exists
    filepath = os.path.join(result_dir, result_filename)
    if not overwrite and os.path.exists(filepath):
        contents = get_hdf_dictionary(result_dir, result_filename)
        if "exp_data" in contents or "exp_metadata" in contents:
            print("File {} already preprocessed, doing nothing.".format(result_filename))
            return
    
    # Prepare the BMI3D data
    bmi3d_data, bmi3d_metadata = parse_bmi3d(data_dir, files)
    if save_res:
        save_hdf(result_dir, result_filename, bmi3d_data, "/exp_data", append=True)
        save_hdf(result_dir, result_filename, bmi3d_metadata, "/exp_metadata", append=True)
        print('done!')
    return bmi3d_data, bmi3d_metadata

def proc_mocap(data_dir, files, result_dir, result_filename, overwrite=False):
    '''
    Process motion capture files:
        Loads metadata, position data, and rotation data from 'optitrack' files
        If present, reads 'hdf' metadata to find appropriate strobe channel
        If present, loads 'ecube' analog data representing optitrack camera strobe
    The data is prepared along with timestamps into HDF datasets:
        mocap_data:
            optitrack [('position', 'f8', (3,)), ('rotation', 'f8', (4,)), ('timestamp', 'f8', (1,)]
        mocap_metadata:
            source_dir (str)
            source_files (dict)
            samplerate (float)
            <other metadata from motive>
    
    Args:
        data_dir (str): where the data files are located
        files (dict): dictionary of filenames indexed by system
        result_filename (str): where to store the processed result
        overwrite (bool): whether to remove existing processed files if they exist

    Returns:
        None
    '''  
    # Check if a processed file already exists
    filepath = os.path.join(result_dir, result_filename)
    if not overwrite and os.path.exists(filepath):
        contents = get_hdf_dictionary(result_dir, result_filename)
        if "mocap_data" in contents or "mocap_metadata" in contents:
            print("File {} already preprocessed, doing nothing.".format(result_filename))
            return

    # Parse Optitrack data
    if 'optitrack' in files:
        optitrack_data, optitrack_metadata = parse_optitrack(data_dir, files)
        save_hdf(result_dir, result_filename, optitrack_data, "/mocap_data", append=True)
        save_hdf(result_dir, result_filename, optitrack_metadata, "/mocap_metadata", append=True)

def proc_eyetracking(data_dir, files, result_dir, exp_filename, result_filename, debug=True, overwrite=False, save_res=True, **kwargs):
    '''
    Loads eyedata from ecube analog signal and calculates calibration profile using least square fitting.
    Requires that experimental data has already been preprocessed in the same result hdf file.
    
    The data is prepared into HDF datasets:
    
    eye_data:
        eye_closed_mask (nt, nch): boolean mask of when the eyes are closed
        raw_data (nt, nch): raw eye data
        calibrated_data (nt, nch): calibrated eye data
        coefficients (nch, 2): linear regression coefficients
        correlation_coeff (nch): best fit correlation coefficients from linear regression
        cursor_calibration_data (ntr, 2): cursor coordinates used for calibration
        eye_calibration_data (ntr, nch): eye coordinates used for calibration
    
    eye_metadata:
        samplerate (float): sampling rate of the calibrated eye data
        see :func:`aopy.preproc.parse_oculomatic` for oculomatic metadata

    Args:
        data_dir (str): where the data files are located
        files (dict): dictionary of filenames indexed by system
        result_dir (str): where to store the processed result 
        result_filename (str): what to call the preprocessed filename
        debug (bool, optional): if true, prints additional debug messages
        overwrite (bool, optional): whether to recalculated and overwrite existing preprocessed eyetracking data
        save_res (bool, optional): whether to save the calculated eyetracking data
        **kwargs (dict, optional): keyword arguments to pass to :func:`aopy.preproccalc_eye_calibration()`

    Returns:
        eye_dict (dict): all the data pertaining to eye tracking, calibration
        eye_metadata (dict): metadata for eye tracking

    Example:
        Uncalibrated raw data:

        .. image:: _images/eye_trajectories_raw.png

        After calibration:
        
        .. image:: _images/eye_trajectories_calibrated.png
    '''
    # Check if data already exists
    filepath = os.path.join(result_dir, result_filename)
    if not overwrite and os.path.exists(filepath):
        contents = get_hdf_dictionary(result_dir, result_filename)
        if "eye_data" in contents and "eye_metadata" in contents:
            print("Eye data already preprocessed in {}, returning existing data.".format(result_filename))
            eye_data = load_hdf_group(result_dir, result_filename, 'eye_data')
            eye_metadata = load_hdf_group(result_dir, result_filename, 'eye_metadata')
            return eye_data, eye_metadata
    
    # Load the preprocessed experimental data
    try:
        exp_data = load_hdf_group(result_dir, exp_filename, 'exp_data')
        exp_metadata = load_hdf_group(result_dir, exp_filename, 'exp_metadata')
    except (FileNotFoundError, ValueError):
        raise ValueError(f"File {exp_filename} does not include preprocessed experimental data. Please call proc_exp() first.")
    
    # Parse the raw eye data; this could be extended in the future to support other eyetracking hardware
    eye_data, eye_metadata = parse_oculomatic(data_dir, files, debug=debug)
    eye_mask = eye_data['mask']
    eye_data = eye_data['data']

    try:
        # Calibrate the eye data
        cursor_samplerate = exp_metadata['cursor_interp_samplerate']
        cursor_data = exp_data['cursor_interp']
        events = exp_data['events']
        event_codes = events['code']
        event_times = events['timestamp'] # time points in the ecube time frame
        coeff, correlation_coeff, cursor_calibration_data, eye_calibration_data = calc_eye_calibration(
            cursor_data, cursor_samplerate, eye_data, eye_metadata['samplerate'], 
            event_times, event_codes, return_datapoints=True, **kwargs)

        calibrated_eye_data = postproc.get_calibrated_eye_data(eye_data, coeff)
        eye_dict = {
            'eye_closed_mask': eye_mask,
            'raw_data': eye_data,
            'calibrated_data': calibrated_eye_data,
            'coefficients': coeff,
            'correlation_coeff': correlation_coeff,
            'cursor_calibration_data': cursor_calibration_data,
            'eye_calibration_data': eye_calibration_data
        }
        try:
            eye_metadata['calibration_version'] = version('aolab-aopy')
        except:
            eye_metadata['calibration_version'] = 'unknown'
        eye_metadata['calibration_date'] = datetime.datetime.now().isoformat()

    except (KeyError, ValueError):
        # If there is no cursor data or there aren't enough trials, this will fail. 
        # We should still save the eye data, just don't include the calibrated data
        eye_dict = {
            'eye_closed_mask': eye_mask,
            'raw_data': eye_data
        }

    # Save everything into the HDF file
    if save_res:
        save_hdf(result_dir, result_filename, eye_dict, "/eye_data", append=True)
        save_hdf(result_dir, result_filename, eye_metadata, "/eye_metadata", append=True)
    return eye_dict, eye_metadata

def proc_broadband(data_dir, files, result_dir, result_filename, overwrite=False, max_memory_gb=1.):
    '''
    Process broadband data:
        Loads 'ecube' headstage data and metadata
    Saves broadband data into the HDF datasets:
        broadband_data (nt, nch)
        broadband_metadata (dict)

    Args:
        data_dir (str): where the data files are located
        files (dict): dictionary of filenames indexed by system
        result_filename (str): where to store the processed result
        overwrite (bool, optional): whether to remove existing processed files if they exist
        max_memory_gb (float, optional): max memory used to load binary data at one time

    Returns:
        None
    '''

    # Check if a processed file already exists
    filepath = os.path.join(result_dir, result_filename)
    if not overwrite and os.path.exists(filepath):
        contents = get_hdf_dictionary(result_dir, result_filename)
        if "broadband_data" in contents:
            raise FileExistsError("File {} already preprocessed, doing nothing.".format(result_filename))
    elif os.path.exists(filepath):
        os.remove(filepath) # maybe bad, since it deletes everything, not just broadband data

    # Copy the broadband data into an HDF dataset
    if 'ecube' in files:
        
        # Process the binary data
        data_filepath = os.path.join(data_dir, files['ecube'])
        result_filepath = os.path.join(result_dir, result_filename)
        _, metadata = proc_ecube_data(data_filepath, 'Headstages', result_filepath, result_name='broadband_data', max_memory_gb=max_memory_gb)

        # Append the broadband metadata to the file
        save_hdf(result_dir, result_filename, metadata, "/broadband_metadata", append=True)

def proc_lfp(data_dir, files, result_dir, result_filename, overwrite=False, max_memory_gb=1., filter_kwargs={}):
    '''
    Process lfp data:
        Loads 'broadband' hdf data or 'ecube' headstage data and metadata
    Saves lfp data into the HDF datasets:
        lfp_data (nt, nch)
        lfp_metadata (dict)
    
    Args:
        data_dir (str): where the data files are located
        files (dict): dictionary of filenames indexed by system
        result_filename (str): where to store the processed result
        overwrite (bool, optional): whether to remove existing processed files if they exist
        max_memory_gb (float, optional): max memory used to load binary data at one time
        filter_kwargs (dict, optional): keyword arguments to pass to :func:`aopy.precondition.filter_lfp`
    '''  
    # Check if a processed file already exists
    filepath = os.path.join(result_dir, result_filename)
    if not overwrite and os.path.exists(filepath):
        contents = get_hdf_dictionary(result_dir, result_filename)
        if "lfp_data" in contents:
            print("File {} already preprocessed, doing nothing.".format(result_filename))
            return
    elif os.path.exists(filepath):
        os.remove(filepath)

    # Preprocess neural data into lfp   
    if 'broadband' in files:
        broadband_filepath = os.path.join(result_dir, files['broadband'])
        result_filepath = os.path.join(result_dir, result_filename)

        _, lfp_metadata = aodata.filter_lfp_from_broadband(broadband_filepath, result_filepath, dtype='int16', 
                                                    max_memory_gb=max_memory_gb, **filter_kwargs)
    elif 'ecube' in files:
        ecube_filepath = os.path.join(data_dir, files['ecube'])
        result_filepath = os.path.join(result_dir, result_filename)

        _, lfp_metadata = aodata.filter_lfp_from_ecube(ecube_filepath, result_filepath, dtype='int16', 
                                                max_memory_gb=max_memory_gb, **filter_kwargs)
        
    save_hdf(result_dir, result_filename, lfp_metadata, "/lfp_metadata", append=True)



import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import butter, filtfilt

from ..precondition import downsample
from ..utils import derivative, detect_edges

def filter_eye(eye_pos, samplerate, downsamplerate=1000, low_cut=200, buttord=4):
    '''
    Filter and downsample eye data.

    .. image:: _images/proc_oculomatic_freq.png

    Args:
        eye_pos (nt, nch): eye position data, e.g. from oculomatic
        samplerate (float): original sampling rate of the eye data
        downsamplerate (float, optional): sampling rate at which to return eye data
        lowpass_freq (float, optional): low cutoff frequency to limit analysis of saccades
        taper_len (float, optional): length of tapers to use in multitaper lowpass filter

    Returns:
        tuple: tuple containing:
        | **eye_pos (nt, nch):** eye position after filtering and downsampling
        | **samplerate (float):** sampling rate of the returned eye data
    '''
    # Lowpass filter
    b, a = butter(buttord, low_cut, btype='lowpass', fs=samplerate)
    eye_pos = filtfilt(b, a, eye_pos, axis=0)

    # Downsample
    if samplerate > downsamplerate:
        eye_pos = downsample(eye_pos, samplerate, downsamplerate)
    else:
        downsamplerate = samplerate

    return eye_pos, downsamplerate

def convert_pos_to_speed(eye_pos, samplerate):
    '''
    Differentiate to get speed from eye position.

    Args:
        eye_pos (nt, nch): eye position data, e.g. from oculomatic
        samplerate (float): sampling rate of the eye data
    
    Returns:
        (nt, nch) array: eye speed

    Examples:

        .. image:: _images/convert_pos_to_accel_nofilter.png

        .. image:: _images/convert_pos_to_accel_filter.png
    '''
    time = np.arange(eye_pos.shape[0])/samplerate
    speed = derivative(time, eye_pos, norm=True)

    return speed    

def convert_pos_to_accel(eye_pos, samplerate):
    '''
    Differentiate twice to get acceleration from eye position.

    Args:
        eye_pos (nt, nch): eye position data, e.g. from oculomatic
        samplerate (float): sampling rate of the eye data
    
    Returns:
        (nt, nch) array: eye acceleration
    '''
    speed = convert_pos_to_speed(eye_pos, samplerate)
    time = np.arange(eye_pos.shape[0])/samplerate
    accel = derivative(time, speed, norm=False)

    return accel

def detect_saccades(eye_pos, samplerate, thr=None, num_sd=1.5, intersaccade_min=None, 
                    min_saccade_duration=0.015, max_saccade_duration=0.16, 
                    lowpass_filter_freq=30, debug=False, debug_window=(0, 2)):
    '''
    Detect saccades from eye kinematics data. Uses thresholding of the kinematics 
    to find putative saccades, then ensures that saccades are spaced by some 
    minimum intersaccade time, and finally only considers events with both positive 
    and negative crossings within the minimum saccade time. Optional plots can be 
    generated showing threshold crossings in position, velocity, and acceleration.

    Minimum and maximum saccade durations can be enforced to remove events that are 
    biologically implausible. Default minimum and maximum saccade durations from 
    Dorr et al., 2010 https://doi.org/10.1167/10.10.28
    
    A minimum intersaccade interval can also be set to prevent the categorization 
    of potential overshoot components as saccadic events. Default intersaccade 
    minimum from Sinn & Engbert, 2016 https://doi.org/10.1016/j.visres.2015.05.012 
    
    Example:

        Detecting saccades on the first 5 seconds of test data from beignet block 5974 (included in /tests/data/beignet)
    
        .. image:: _images/detect_saccades.png
        .. image:: _images/detect_saccades_hist.png
        .. image:: _images/detect_saccades_scatter.png
    
    Args:
        eye_kinematics (nt, nch): eye kinematics data, e.g. velocity or acceleration for one eye
        samplerate (float): sampling rate of the eye data
        thr ((high,low) tuple): positive and negative threshold values for acceleration, if desired. If None (default), 
            thresholds will be automatically chosen based on num_sd above the mean.
        num_sd (float, optional): number of standard deviations above zero to threshold acceleration
        intersaccade_min (float, optional): minimum time (in seconds) allowed between saccades (from offset to next onset)
        min_saccade_duration (float, optional): minimum time (in seconds) that a saccade can take (inclusive)
        max_saccade_duration (float, optional): maximum time (in seconds) that a saccade can take (exclusive)
        lowpass_filter_freq (float, optional): frequency in Hz to low-pass filter. If None, no filtering will take place. Default 30 Hz. 
        debug (bool, optional): if True, display a figure showing the threshold crossings
        debug_window (tuple, optional): (start, end) time (in seconds) for the debug figure
        
    Returns:
        tuple: tuple containing:
            | **onset (nsaccade):** onset time (in seconds) of each detected saccade
            | **duration (nsaccade):** duration (in seconds) of each detected saccade
            | **distance (nsaccade):** distance (same units as eye_pos) of each detected saccade

    '''
    assert thr is None or (len(thr) == 2 and thr[0] > thr[1]), ("Threshold must be in the form" 
        " (positive thr, negative thr)")
    assert (intersaccade_min is None) or intersaccade_min < max_saccade_duration, ("Max saccade"
        " duration must be longer than the minimum intersaccade interval")
    
    if lowpass_filter_freq is not None:
        eye_pos, _ = filter_eye(eye_pos, samplerate, samplerate, low_cut=lowpass_filter_freq)    
    eye_accel = convert_pos_to_accel(eye_pos, samplerate)

    # Set an appropritate threshold to detect saccades
    if thr is None:
        baseline_mean = np.mean(eye_accel)
        baseline_std = np.std(eye_accel)
        thr = np.mean(baseline_mean) + num_sd*baseline_std
        thr = (thr, -thr)
    saccade_onset_time, _ = detect_edges(eye_accel > thr[0], samplerate, rising=True, falling=False)
    saccade_offset_time, _ = detect_edges(eye_accel < thr[1], samplerate, rising=False, falling=True)

    if saccade_onset_time.size == 0 or saccade_offset_time.size == 0:
        return [], [], []

    # Only consider saccades that have a threshold-crossing offset
    onset = []
    duration = []
    distance = []
    prev_offset = 0
    for t in saccade_onset_time:
        if intersaccade_min and t < prev_offset + intersaccade_min:
            continue # ignore saccades within intersaccade_min of the last saccade
        nearby_offset = np.where((saccade_offset_time >= t + min_saccade_duration) & 
                                 (saccade_offset_time < t + max_saccade_duration))[0]
        if len(nearby_offset):
            offset_time = saccade_offset_time[nearby_offset[0]]
            onset.append(t)
            duration.append(offset_time - t)
            distance.append(
                np.linalg.norm(eye_pos[int(t*samplerate),:] - 
                               eye_pos[int(offset_time*samplerate),:])
            ) # distance is the deviation of the saccade in space (in cm)
            prev_offset = offset_time
    onset = np.array(onset)
    duration = np.array(duration)
    distance = np.array(distance)

    if debug:
        time = np.arange(len(eye_pos))/samplerate
        debug_window = (debug_window[0], min(debug_window[1], time[-1]))
        debug_idx = (int(debug_window[0]*samplerate), int(debug_window[1]*samplerate))
        fig, ax = plt.subplots(2,1)
        ax[0].plot(time[debug_idx[0]:debug_idx[1]], eye_pos[debug_idx[0]:debug_idx[1]])
        ax[1].plot(time[debug_idx[0]:debug_idx[1]], eye_accel[debug_idx[0]:debug_idx[1]])
        ax[0].set_ylabel('position (cm)')
        ax[1].set_xlabel('time (s)')
        ax[1].set_ylabel('accel cm/s^2')
        ax[1].plot([debug_window[0],debug_window[1]], [thr[0], thr[0]], '--')
        ax[1].plot([debug_window[0],debug_window[1]], [thr[1], thr[1]], '--')
        
        min_max = [np.min(eye_pos[debug_idx[0]:debug_idx[1]]), np.max(eye_pos[debug_idx[0]:debug_idx[1]])]
        debug_idx = (onset > debug_window[0]) & (onset < debug_window[1])
        onset_debug = onset[debug_idx]
        duration_debug = duration[debug_idx]
        for o, d in zip(onset_debug, duration_debug):
            ax[0].plot([o, o], min_max, 'g--')
            ax[0].plot([o+d, o+d], min_max, 'r--')
            ax[1].plot([o, o], [thr[0], thr[1]], 'g--')
            ax[1].plot([o+d, o+d], [thr[0], thr[1]], 'r--')

    return onset, duration, distance
        



# precondition.py
# Code for cleaning and preparing neural data for users to interact with,
# for example: down-sampling, outlier detection, and initial filtering

from os import fsdecode, fspath, fstat
import warnings
from scipy import signal
from scipy.signal import butter, lfilter, filtfilt, decimate, windows
import numpy as np
import math

from .. import analysis
from .. import utils
'''
Filter functions
'''

def butterworth_params(cutoff_low, cutoff_high, fs, order = 4, filter_type = 'bandpass'):
    '''
    Design Nth-order digital Butterworth filter and return the filter coefficients.

    Args:
        cutoff_low (int): lower cut-off frequency (in Hz)
        cutoff_high (int): higher cutoff frequency (in Hz)
        fs (int): sampling rate (in Hz)
        order (int): Order of the butter worth filter
        filter_type (str) : Type of filter. Accepts one of the four values - {`lowpass`, `highpass`, `bandpass`, `bandstop`}

    Returns:
        tuple (b,a): bandpass filter parameters

    '''

    if filter_type == 'lowpass':
        Wn = cutoff_high

    if filter_type == 'highpass':
        Wn = cutoff_low

    if filter_type in ['bandpass', 'bandstop']:
        Wn = [cutoff_low, cutoff_high]

    b,a = butter( order, Wn, btype=filter_type, fs =fs)
    return b,a

def butterworth_filter_data(data, fs, cutoff_freqs=None, bands=None, order=None, filter_type='bandpass'):
    '''
    Apply a digital butterworth filter forward and backward to a timeseries signal.

    Args:
        data (nt, ...): neural data
        fs (int): sampling rate (in Hz)
        cutoff_freqs (float): list of cut-off frequencies (in Hz); only for 'high pass' or 'low pass' filter. Use bands for 'bandpass' filter
        bands (list): frequency bands should be a list of tuples representing ranges e.g., bands = [(0, 10), (10, 20), (130, 140)] for 0-10, 10-20, and 130-140 Hz
        order (int): Order of the butterworth filter. If no order is specified, the function will find the minimum order of filter required to maintain +3dB gain in the bandpass range.
        filter_type (str) : Type of filter. Accepts one of the four values - {'lowpass', 'highpass', 'bandpass', 'bandstop'}

    Returns:
        tuple: Tuple containing:
            | **filtered_data (list):** output bandpass filtered data in the form of a list. Each list item is filtered data in corresponding frequency band
            | **Wn (list):** frequency bands
    '''

    if filter_type in ['lowpass', 'highpass']:
        Wn = cutoff_freqs

    if filter_type in ['bandpass', 'bandstop']:
        if not bands:
            raise ValueError("Must provide lower and higher cut off frequency")

        Wn = np.zeros((len(bands),2))
        for ib, band in enumerate(bands):
            Wn[ib] = [band[0], band[1]]

    filtered_data = []
    for _, w in enumerate(Wn):

        # Automatically select an order if none is specified
        if order is None:
            wp = w/(2 * fs)
            if type(wp) != int and type(wp) != float and len(wp) == 2:
                ws = [wp[0] - 0.1 , wp[1] + 1]
            else:
                ws = wp - 0.1
            order, _ = signal.buttord(wp, ws, 3, 40, False)

        # Filter this frequency or band
        b, a = butter(order, w, btype=filter_type, fs= fs)
        # filtered_data = lfilter(b,a,data)
        filtered_data.append(filtfilt(b, a, data, axis=0))
    return filtered_data, Wn

'''
Filter functions related to multitaper method
'''
def dpsschk(n, p, k):
    '''
    Computes the Discrete Prolate Spheroidal Sequences (DPSS) array based on input tapers

    Args:
        n (float): window length in number of samples
        p (float): standardized half bandwidth in hz
        k (int): number of DPSS tapers to use

    Returns:
        e ((n, k) array): K DPSS windows
        v ((k,) array): The concentration ratios for the K windows
    '''
    n = round(n)

    if k < 1:
        raise Exception('Error:  K must be greater than or equal to 1')
    elif k > 2*p-1:
        raise Exception('Error:  K must be less than 2*P-1')

    e, v = windows.dpss(n, p, k,return_ratios=True)
    e = e.T
    
    return e, v

def dp_proj(tapers, fs=1, f0=0):
    '''
    Generates a prolate projection operator

    Args:
        tapers ((n, k) array): data tapers
        fs (float): sampling rate
        f0 (float): center frequency of filiter

    Returns:
        dp_proj (nt, K): projection operator in [time, K] form
    '''
    N = tapers.shape[0]
    K = tapers.shape[1]
    pr_op = np.zeros((N,K),dtype = 'complex')

    shifter = np.exp(-2.*np.pi*1j*f0*np.arange(1,N+1)/fs)
    for k in range(K):
        pr_op[:,k] = shifter.conj()*tapers[:,k]
        
    return pr_op

def compute_bp_filter(tapers, fs=1, f0=0):
    '''
    Generates a bandpass filter using the multitaper method

    Args:
        tapers ((n, k) array): data tapers
        fs (float): sampling rate
        f0 (float): center frequency of filiter

    Returns:
        X (1, 2*N*fs): a bandpass filter
    '''    
    pr_op = dp_proj(tapers, fs, f0)
    N = pr_op.shape[0]
    X = np.zeros( (1,2*N), dtype = 'complex')
    pr = pr_op@pr_op.conj().T

    for t in range(N):
        X[0, t:t+N] = X[0, t:t+N] + pr[:,N-1-t].conj().T
    X = X/N 
    
    return X

def convert_taper_parameters(n, w):
    '''        
    Converts tapers in [n, w] form to [n, p, k] form. It's wise to print out 
    the number of tapers (k) to check that there will be enough frequency 
    resoltuion in the filter.
    P is computed by N*W and K is given by math.floor(2*P-1)

    Args:
        n (float): time window for analysis (in seconds)
        w (float): half bandwith of tapers (in hz)

    Returns:
        tuple: [n, p, k] taper parameters
    '''
    p = n*w
    k = math.floor(2*p-1)
    if k < 1:
        warnings.warn("K is less than 1, try increasing the taper length or the half-bandwidth")
    return n, p, k

def mt_lowpass_filter(data, lowcut, taper_len, fs, verbose=True, **kwargs):
    '''
    Wrapper around mtfilter that auto-generates [n, p, k] and f0 parameters 
    based on the lowpass of interest, length of the tapers, and the sampling 
    rate. See :func:`~aopy.precondition.mtfilter` for more details.

    Args:
        data (nt, nch): time series array
        lowcut (float): low-pass cutoff frequency
        taper_len (float): length of the tapers (in seconds) 
        fs (float): sampling rate of the data
        verbose (bool, optional): if True, print out the n, k parameters.
        kwargs (dict): additional keyword-arguments to pass to mtfilter

    Returns:
        (nt, nch): filtered time-series data
    '''
    n, p, k = convert_taper_parameters(taper_len, lowcut)
    f0 = 0
    if verbose:
        print(f"Using {k} tapers of length {n:.4f} s")
    return mtfilter(data, n, p, k, fs, f0, **kwargs)


def mt_bandpass_filter(data, band, taper_len, fs, verbose=True, **kwargs):
    '''
    Wrapper around mtfilter that auto-generates [n, p, k] and f0 parameters based on the band of interest,
    length of the tapers, and the sampling rate. See :func:`~aopy.precondition.mtfilter` for more details.

    Args:
        data (nt, nch): time series array
        band ((2,) tuple): low and high frequencies to band-pass
        taper_len (float): length of the tapers (in seconds) 
        fs (float): sampling rate of the data
        verbose (bool, optional): if True, print out the n, k parameters.
        kwargs (dict): additional keyword-arguments to pass to mtfilter

    Returns:
        (nt, nch): filtered time-series data

    '''
    n, p, k = convert_taper_parameters(taper_len, (band[1]-band[0])/2)
    f0 = np.mean(band)
    if verbose:
        print(f"Using {k} tapers of length {n:.4f} s")
    return mtfilter(data, n, p, k, fs, f0, **kwargs)


def mtfilter(data, n, p, k, fs=1, f0=0, center_output=True, complex_output=False, use_fft=True):
    '''
    Bandpass-filter a time series data using the multitaper method. Tapers are calculated using [n, p, k]
    parameters (see :func:`~aopy.precondition.convert_taper_parameters`). Be careful to set these parameters
    to control temporal and spectral properties of the filter. The parameters you pick will depend on the 
    time-scale of questions you're interested in. You get better estimates by smoothing across more time, 
    but at the cost of temporal resolution. For example, if responses that happen quickly, you'll want to keep 
    the taper length relatively small whenever possible. But that then blurs your frequency resolution, 
    so you need to play around with the bandwidth and taper length to find the lowest temporal resolution 
    you can get away with for a given frequency range goal.

    Example:
        ::

            band = [575, 625] # signals within band can pass
            N = 0.1
            NW = (band[1]-band[0])/2
            f0 = np.mean(band)
            n, p, k = convert_taper_parameters(N, NW)
            print(f"Using {k} tapers of half bandwidth {p:.2f}")

            T = 0.05
            fs = 25000
            x, t = utils.generate_test_signal(T, fs, freq=[600, 312, 2000], a=[5, 2, 0.5], noise=0.2)

            x_mtfilter = precondition.mtfilter(x, tapers, fs=fs, f0=f0)

        .. image:: _images/mtfilter.png

    Args:
        data (nt, nch): time series array
        n (float): window length in seconds
        p (float): standardized half bandwidth in hz
        k (int): number of DPSS tapers to use
        fs (float): sampling rate
        f0 (float, optional): center frequency of filiter. Default 0.
        center_output (bool, optional): if True, output data is centered. Default True.
        complexflag (bool, optional): if False, output data becomes real. Default False.
        use_fft (bool, optional): if True, use FFT to convolve data with filter, should be faster. Default True.

    Returns:
        (nt, nch): filtered time-series data
    '''   
    tapers, _ = dpsschk(n*fs, p, k)
        
    if data.ndim == 1:
        data = np.reshape(data,(-1,1))
        
    filt = compute_bp_filter(tapers, fs, f0)
    N = filt.shape[1]
    sz = data.shape

    if complex_output:
        Y = np.zeros(sz) + 1j*np.zeros(sz)
    else:
        filt = filt.real
        Y = np.zeros(sz)

    if use_fft:
        tmp = signal.fftconvolve(data, filt.T, axes=0)
        if center_output:
            Y = tmp[round(N/2):sz[0]+round(N/2), :]
        else:
            Y = tmp[N-1:sz[0]+N,:]
    else:
        for ii in range(sz[1]):
            tmp = np.convolve(data[:,ii], filt[0,:].T)
            if center_output:
                Y[:,ii] = tmp[round(N/2):sz[0]+round(N/2)]
            else:
                Y[:,ii] = tmp[N-1:sz[0]+N]

    if f0 > 0:
        Y = 2*Y
    
    return Y


'''
Spike detection functions
'''

def calc_spike_threshold(spike_filt_data, high_threshold=True, rms_multiplier=3):
    '''
    Use the RMS of each channel to set a different threshold for each channel. Sadtler et al. 2014 set threshold to 3.0x RMS value for each channel, which is the default for this function.
    The threshold value will be calculated with the mean subtracted, then the mean for each signal will be added back to the threshold value.
    
    Args:
        spike_filt_data (nt, ...): Filtered time series data.
        high_threshold (bool): If the threshold should allow spikes to be detected above the threshold (True) or below the threshold (False). Defaults to true.
        rms_multiplier (float): Value to multiply the rms value of each time series by.

    Returns:
        threshold_values: Threshold values along the first axis. Output dimensions will be the same non-time dimensions as the input signal.
    
    '''
    mean_input_data = np.mean(spike_filt_data, axis=0)
    rms_values = analysis.calc_rms(spike_filt_data, remove_offset=True)

    if high_threshold:
        return (rms_multiplier*rms_values)+mean_input_data
    else:
        return mean_input_data-(rms_multiplier*rms_values)


def detect_spikes(spike_filt_data, samplerate, threshold, above_thresh=True,  wf_length=1000, tbefore_wf=200):
    '''
    This function calculates spike times based on threshold crossing of the input data and returns the waveforms if 'wf_length' is not None. 
    If the threshold desired is a negative value (i.e. extracellular recordings) set 'above_thresh' to False. 
    Data must exceed the threshold instead of equaling it. 

    Args:
        spike_filt_data (nt, nch): Time series neural data to detect spikes and extract waveforms from.
        samplerate (float): Sampling rate [Hz]
        threshold (nch): Threshold that input data must cross to indicate a spike for each channel. Must have same non time dimensions as spike_filt_data. 
        above_thresh (bool): If True, only spikes above the threshold will be detected. If false, only spikes below threshold will be detected. 
        wf_length (float): Length of waveforms to output :math:`[\mu s]`. Actual length will be rounded up. If set to 'None', waveforms will not be returned. Defaults to 1000 :math:`\mu s`
        tbefore_wf (float): Length of waveform to include before spike detection time :math:`[\mu s]`. Defaults to 200 :math:`\mu s`  
        
    Returns: 
        tuple: Tuple containing:
            | **spike_times (list of spike times):**  List of nspike length arrays with each list element corresponding to a channel. Spike times are in seconds.
            | **spike_waveforms (list of waveforms):** List of (nspike, nwf_pts) arrays with each list element corresponding to a channel. Returns NaN if there aren't enough data points to retreive a full waveform.
    '''
    if spike_filt_data.ndim == 1:
        spike_filt_data = spike_filt_data.reshape(-1,1)
    nch = spike_filt_data.shape[1]

    # Calculate an array of spike times for each channel organized into a list
    if above_thresh:        
        data_above_thresh_mask = spike_filt_data > threshold
        
    elif above_thresh == False:        
        data_above_thresh_mask = spike_filt_data < threshold

    spike_times = []
    spike_waveforms =[]
    
    for ich in range(nch):
        # Spike times
        all_spike_times, _ = utils.detect_edges(data_above_thresh_mask[:,ich], samplerate, rising=True, falling=False)
        spike_times.append(all_spike_times)


        # Spike waveforms
        if wf_length is not None:
            wf_idx_length = math.ceil(samplerate*(wf_length*(1e-6)))
            nspikes = len(all_spike_times)

            temp_spike_waveforms = np.zeros((nspikes, wf_idx_length))
            temp_spike_waveforms[:] = np.nan

            for ispike in range(nspikes):
                starttime = all_spike_times[ispike]-(tbefore_wf*(1e-6))
                startidx = int(np.round(starttime*samplerate, decimals=0))
                stopidx = startidx + wf_idx_length

                if np.logical_and(stopidx < spike_filt_data.shape[0], startidx>=0): # Ensure there are enough data points to grab the waveform
                    temp_spike_waveforms[ispike, :] = spike_filt_data[startidx:stopidx,ich]
        
            spike_waveforms.append(temp_spike_waveforms)

    return spike_times, spike_waveforms

def detect_spikes_chunk(spike_filt_data, samplerate, threshold, chunksize, above_thresh=True,  wf_length=1000, tbefore_wf=200):
    '''
    This function is based on detect_spikes and calculates spike times while dividing data into chunks to deal with memory issues.
    This may work for numpy memmap array

    Args:
        spike_filt_data (nt,nch): Time series neural data to detect spikes and extract waveforms from.
        samplerate (float): Sampling rate [Hz]
        threshold (float): Threshold that input data must cross to indicate a spike for each channel.
        chunksize (int): Chunk size to process data by chunk
        above_thresh (bool): If True, only spikes above the threshold will be detected. If false, only spikes below threshold will be detected. 
        wf_length (float): Length of waveforms to output :math:`[\mu s]`. Actual length will be rounded up. If set to 'None', waveforms will not be returned. Defaults to 1000 :math:`\mu s`
        tbefore_wf (float): Length of waveform to include before spike detection time :math:`[\mu s]`. Defaults to 200 :math:`\mu s`  
        
    Returns: 
        tuple: Tuple containing:
            | **spike_times (list of spike times):**  List of nspike length arrays with each list element corresponding to a channel. Spike times are in seconds.
            | **spike_waveforms (list of waveforms):** List of (nspike, nwf_pts) arrays with each list element corresponding to a channel. Returns NaN if there aren't enough data points to retreive a full waveform.
    '''
    if spike_filt_data.ndim == 1:
        spike_filt_data = spike_filt_data.reshape(-1,1)
        
    nt,nch = spike_filt_data.shape
    nchunk = math.ceil(nt/chunksize)
    chunktime = chunksize/samplerate
    
    # Calculate spike times
    n_samples = 0
    previous_thresh = np.ones(nch) == 1
    spike_times = []
    spike_waveforms =[]
    
    for ichunk in range(nchunk):
        if ichunk == nchunk:
            data = spike_filt_data[n_samples:,:]
        else:
            data = spike_filt_data[n_samples:n_samples+chunksize,:]

        if above_thresh:
            data_above_thresh_mask = data > threshold
        else:        
            data_above_thresh_mask = data < threshold

        # Spike times at each chunk (temp_spike_times: a list of array with different shape)

        for ich in range(nch):
            spike_times_chunk, _ = utils.detect_edges(data_above_thresh_mask[:,ich], samplerate, rising=True, falling=False)
            spike_times_chunk = spike_times_chunk + ichunk*chunktime # Take into account chunk time
            
            if ichunk == 0:
                spike_times.append(spike_times_chunk)
            else:
                if not previous_thresh[ich]: # Check if the last data of the previous chunk is below threshold
                    if data_above_thresh_mask[0,ich]:
                        spike_times_chunk = np.append(ichunk*chunktime, spike_times_chunk)   
                        
                spike_times[ich] = np.append(spike_times[ich],spike_times_chunk)
            
            # Check if data is below threshold at the last point of the chunk
            if data_above_thresh_mask[-1,ich]:
                previous_thresh[ich] = True
            else:
                previous_thresh[ich] = False # This is for next chunk
                
            # Spike waveforms at each chunk and channel
            if wf_length is not None:
                wf_idx_length = math.ceil(samplerate*(wf_length*(1e-6)))
                nspikes = len(spike_times_chunk)

                spike_waveforms_chunk = np.zeros((nspikes, wf_idx_length))
                spike_waveforms_chunk[:] = np.nan

                for ispike in range(nspikes):
                    starttime = spike_times_chunk[ispike]-(tbefore_wf*(1e-6))
                    startidx = int(np.round(starttime*samplerate, decimals=0))
                    stopidx = startidx + wf_idx_length

                    if np.logical_and(stopidx < nt, startidx>=0): # Ensure there are enough data points to grab the waveform
                        spike_waveforms_chunk[ispike, :] = spike_filt_data[startidx:stopidx,ich]
                
                if ichunk == 0:
                    spike_waveforms.append(spike_waveforms_chunk)
                else:
                    spike_waveforms[ich] = np.concatenate([spike_waveforms[ich],spike_waveforms_chunk],axis=0)
            
        n_samples += chunksize

    return spike_times, spike_waveforms

def filter_spike_times_fast(spike_times, refractory_period=100):
    '''
    This function takes spike times and filters them to remove spikes within the refractory period of the preceding spike. This function always assumes the first threshold crossing is a good spike.
    Note: This function will remove spikes that are in the refractory period of the preceding spike even if the the preceding spike is in the refractory period of its preceding spike even though that spike shouldn't be removed.
    For example if the refractory period is set to 100 :math:`\mu s` and there are spikes at 50 :math:`\mu s`, 125 :math:`\mu s`, and 200 :math:`\mu s`, the spikes at 125 :math:`\mu s` and 200 :math:`\mu s` will be removed even though only the 125 :math:`\mu s` spike should be removed.
    If the refractory period is small enough this shouldn't be a problem. For more thorough spike time filter use aopy.precondition.filter_spike_times although it takes ~50x longer.

    Args:
        spike_times (nspikes): 1D array of spike times.
        refractory_period (float): Length of time after a spike is detected before another spike can be detected :math:`[\mu s]`. Defaults to 100 :math:`\mu s`.

    Returns: 
        tuple: A tuple containing:
            | **filtered_spike_times (ngoodspikes):** Spikes times after preliminary filtering to remove spikes within the refractory period of the preceding spike.
            | **filtered_spike_idx (ngoodspikes):** Indices corresponding to accepted spike times
    '''
    filtered_spike_idx = np.where(np.diff(spike_times, prepend=0) > refractory_period*(1e-6))[0]

    # Assume first spike is always good. If it isn't included, add it back in
    if len(np.where(filtered_spike_idx==0)[0])==0:
        filtered_spike_idx = np.insert(filtered_spike_idx,0,0)

    filtered_spike_times = spike_times[filtered_spike_idx]

    return filtered_spike_times, filtered_spike_idx

def filter_spike_times(spike_times, refractory_period=100):
    '''
    This function takes spike times and filters them to remove spikes within the refractory period of the preceding spike. This function always assumes the first threshold crossing is a good spike.
    This function works by jumping to and recording the next spike time after the refractory period ends from the current spike.

    Args:
        spike_times (nspikes): 1D array of spike times.
        refractory_period (float): Length of time after a spike is detected before another spike can be detected :math:`[\mu s]`. Defaults to 100 :math:`\mu s` .

    Returns: 
        tuple: A tuple containing:
            | **filtered_spike_times (ngoodspikes):** Spikes times after preliminary filtering to remove spikes within the refractory period of the preceding spike.
            | **filtered_spike_idx (ngoodspikes):** Indices corresponding to accepted spike times
    '''

    nspikes = len(spike_times)
    filtered_spike_idx = np.zeros(nspikes, dtype=int)
    filtered_spike_idx[0] = spike_times[0] # Store first spike time
    
    ispike = 0
    counter = 0
  
    while ispike <= nspikes:
        
        # Store current spike time
        current_spike_time = spike_times[ispike]
        filtered_spike_idx[counter] = ispike

        # Get next acceptable spike time and idx
        remaining_spikes = np.where(spike_times > current_spike_time+(refractory_period*(1e-6)))[0]
        
        # If there is no next spike, exit loop, else store it.
        if len(remaining_spikes) == 0:
          break
        else:
          ispike = remaining_spikes[0]

        counter += 1

    # Remove trailing zeros
    filtered_spike_idx = np.trim_zeros(filtered_spike_idx, 'b')
    
    # Get filtered spike times
    filtered_spike_times = spike_times[filtered_spike_idx]

    return filtered_spike_times, filtered_spike_idx


def bin_spikes(data, fs, bin_width):
    '''
    Computes binned spikes [spikes/s]. The input data is the sampled thresholded data (0 or 1 data).
    Binned spikes are calculated at each bin whose width is determined by bin_width. 

    Example:
        >>> data = np.array([[0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1],[1, 1, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 1, 0]])
        >>> data_T = data.T
        >>> fs = 100
        >>> binned_spikes = precondition.bin_spikes(data_T, fs, .05)
        >>> print(binned_spikes)
            [[200. 300.]
            [200. 200.]
            [300. 200.]
            [100. 100.]]

    Args:
        data (nt, nch): time series spike data with multiple channels.
        fs (float): sampling rate of data [Hz]
        bin_width (float): [s] Spikes are summed within 'bin_width' then devided by 'fs'

    Returns:
        binned_spikes (nbin, nch): binned spikes [spikes/s].
    '''

    dT = 1/fs
    bin_width_idx = round(bin_width*fs) # from [s] to index
    nbins = math.ceil(data.shape[0]/bin_width_idx) # the number of bins
    nch = data.shape[1]
    binned_spikes = np.zeros((nbins,nch))

    for idx in range(nbins):
        if idx == nbins - 1:
            binned_spikes[idx,:] = np.sum(data[idx * bin_width_idx :, :], axis = 0)
        else:
            binned_spikes[idx,:] = np.sum(data[idx * bin_width_idx : (idx + 1) * bin_width_idx, :], axis = 0)

    binned_spikes = binned_spikes/bin_width # convert from [spikes/bin] to [spikes/s]    
    return binned_spikes

def calc_ks_waveforms(raw_data, sample_rate, spike_times_unit, templates, channel_pos, waveforms_nch=10, time_before=1000., time_after=1000.):
    '''
    Calculate waveforms, waveform channels, and positions of units, using templates from kilosort
    This function does not account for drift correction
    
    Args:
        raw_data (nt,nch): time series neural data to detect spikes and extract waveforms from.
        sample_rate (float): sampling rate (Hz)
        spike_times_unit (dict): spike times for each unit identified by kilosort
        templates (n_unit, n_points, nch): templates that kilosort used to detect spikes
        channel_pos (nch, 2): channel positions
        waveforms_nch (int, optional): the number of channels with large amplitude of templates
        time_before (float, optional): time (us) to include before the start of each trial
        time_after (float, optional): time (us) to include after the start of each trial
    
    Returns:
        tuple: tuple containing:
            | **unit_waveforms (dict):** waveforms for each unit. The shape is (nspikes,  m_points, waveforms_nch)
            | **unit_waveforms_ch (n_unit, waveforms_nch):** large amplitude channels in templates
            | **unit_pos (dict):** channel positions of each unit
    '''
    
    time_before *= 1e-6
    time_after *= 1e-6
    nch = channel_pos.shape[0]
    duration = int((time_before + time_after)*sample_rate)
    
    unit_waveforms_ch = {}
    unit_waveforms = {}
    unit_pos = {}

    for iunit, unit in enumerate(spike_times_unit.keys()):
        # Look at high amplitude channels in templates
        amp_template_ch = np.zeros(nch)
        for ich in range(nch):
            amp_template_ch[ich] = np.max(templates[int(unit),:,ich])-np.min(templates[int(unit),:,ich]) # don't use iunit instead of int(unit)

        # Sort high amplitude channels and save channels and their positions
        large_amp_ch = np.argsort(amp_template_ch)[::-1][:waveforms_nch]
        unit_waveforms_ch[f'{unit}'] = large_amp_ch
        unit_pos[f'{unit}'] = channel_pos[large_amp_ch[0],:]

        # Get waveforms in high amplitude channels for each spike
        unit_times = spike_times_unit[f'{unit}']
        waveforms = np.zeros((unit_times.shape[0],duration,waveforms_nch))*np.nan
        for ispike, unit_time in enumerate(unit_times):
            start = int((unit_time - time_before)*sample_rate)
            end = start + duration

            if np.logical_and(end < raw_data.shape[0], start >= 0): # Ensure there are enough data points to grab the waveform
                for ich, ch in enumerate(large_amp_ch):
                    waveforms[ispike,:,ich] = raw_data[start:end,ch]
                    
        unit_waveforms[f'{unit}'] = waveforms
        
    return unit_waveforms, unit_waveforms_ch, unit_pos

def bin_spike_times(spike_times, time_before, time_after, bin_width):
    '''
    Computes binned spikes (spike rate) [spikes/s]. The input data are 1D spike times in seconds.
    Binned spikes are calculated at each bin whose width is determined by bin_width. 

    Example:
        >>> spike_times = np.array([0.0208, 0.0341, 0.0347, 0.0391, 0.0407])
        >>> spike_times = spike_times.T
        >>> time_before = 0
        >>> time_after = 0.05
        >>> bin_width = 0.01
        >>> binned_unit_spikes, time_bins = precondition.bin_spike_times(spike_times, time_before, time_after, bin_width)
        >>> print(binned_unit_spikes)
            [  0.   0. 100. 300. 100.]
        >>> print(time_bins)
            [0.005 0.015 0.025 0.035 0.045]

    Args:
        spike_times (nspikes): 1D array of spike times [s]
        time_before (float): start time to easimate spike rate [s]
        time_after (float): end time to estimate spike rate (Estimation includes endpoint)[s]
        bin_width (float): width of time-bin to use for estimating spike rate [s]

    Returns:
        binned_unit_spikes (nbin, nch): spike rate [spikes/s].
        time_bins : the center of the time-bin over which firing rate is estimated. [s]
    '''

    time_bins = np.arange(time_before, time_after+bin_width, bin_width) # contain endpoint

    binned_unit_spikes, _ = np.histogram(spike_times, bins=time_bins)
    binned_unit_spikes = binned_unit_spikes/bin_width # convert [spikes] to [spikes/s]

    time_bins = time_bins[0:-1] + np.diff(time_bins)/2 #change time_bins to be the center of the bin, not the edges.

    return binned_unit_spikes, time_bins

def downsample(data, old_samplerate, new_samplerate):
    '''
    Downsample by averaging. Computes a downsample factor based on old_samplerate/new_samplerate.
    If the downsample factor is fractional, then first upsamples to the least common multiple of 
    the two sampling rates. Finally, pads data to be a multiple of the downsample factor and 
    averages blocks into the new samples. 

    .. image:: _images/downsample.png

    Args:
        data (nt, ...): timeseries data to be downsampled. Can be 1D or 2D.
        old_samplerate (int): the current sampling rate of the data
        new_samplerate (int): the desired sampling rate of the downsampled data
        
    Returns:
        (nt, ...) downsampled data
    '''
    assert new_samplerate < old_samplerate, "New sampling rate must be less than old sampling rate"
    assert int(old_samplerate) == old_samplerate, "Input samplerates must be integers"
    assert int(new_samplerate) == new_samplerate, "Input samplerates must be integers"

    # Check if the downsample factor will be an integer, otherwise we find a common divisor
    if old_samplerate % new_samplerate != 0:
        lcm = np.lcm(int(old_samplerate), int(new_samplerate)) # least common multiple
        print(f"Upsampling first to {lcm} Hz")
        upsampled = np.repeat(data, lcm/old_samplerate, axis=0)
        return downsample(upsampled, lcm, new_samplerate)
        
    old_samples = data.shape[0]
    downsample_factor = int(old_samplerate/new_samplerate)

    # Pad the data to a multiple of the downsample factor
    pad_size = math.ceil(float(old_samples)/downsample_factor)*downsample_factor - old_samples
    pad_shape = (pad_size,)
    if data.ndim > 1:
        pad_shape = np.concatenate(([pad_size], data.shape[1:]))
    data_padded = np.append(data, np.zeros(pad_shape)*np.NaN, axis=0)

    # Downsample using average
    if data.ndim == 1:
        return np.nanmean(data_padded.reshape(-1, downsample_factor), axis=1)
    elif data.ndim == 2:
        downsampled = np.zeros((int(data_padded.shape[0] / downsample_factor), *data.shape[1:]), dtype=data.dtype)
        for idx in range(data.shape[1]):
            downsampled[:, idx] = np.nanmean(data_padded[:, idx].reshape(-1, downsample_factor), axis=1)
        return downsampled
    elif data.ndim == 3:
        downsampled = np.zeros((int(data_padded.shape[0] / downsample_factor), *data.shape[1:]), dtype=data.dtype)
        for idx1 in range(data.shape[1]):
            for idx2 in range(data.shape[2]):
                downsampled[:, idx1, idx2] = np.nanmean(data_padded[:, idx1, idx2].reshape(-1, downsample_factor), axis=1)
        return downsampled

def filter_lfp(broadband_data, broadband_samplerate, lfp_samplerate=1000., low_cut=500., buttord=4):
    '''
    Low-pass filter and return downsampled version of broadband data at default 1000 Hz. Default
    filter parameters taken from Stavisky, Kao, et al., 2015

    Args:
        broadband_data (nt, ...): raw headstage data, e.g.
        broadband_samplerate (float): sampling rate of the raw data
        lfp_samplerate (float, optional): sampling rate of the LFP data. Defaults to 1000.
        low_cut (float, optional): cutoff frequency for low-pass filter. Defaults to 500 Hz
        buttord (int, optional): order for butterworth low-pass filter. Defaults to 4.

    Returns:
        tuple: tuple containing:
        | **lfp_data (nt', ...):** downsampled filtered lfp data
        | **samplerate (float):** sampling rate of the lfp data
    '''
    b, a = butter(buttord, low_cut, btype='lowpass', fs=broadband_samplerate)
    filtered_data = filtfilt(b, a, broadband_data, axis=0)
    return downsample(filtered_data, broadband_samplerate, lfp_samplerate), lfp_samplerate

def filter_spikes(broadband_data, samplerate, low_pass=500, high_pass=7500, buttord=3):
    '''
    Band-pass filter broadband data at default 500-7500 Hz. Default filtering parameters taken
    from Stavisky, Kao, et al., 2015

    Args:
        broadband_data (nt, ...): raw headstage data, e.g.
        samplerate (float): sampling rate of the raw data
        low_pass (float, optional): low-cut frequency, default 500 Hz
        high_pass (float, optional): high-cut frequency, default 7500 Hz
        buttord (int, optional): order for butterworth band-pass filter. Default 3

    Returns:
        (nt, ...): spike filtered data
    '''
    window = [low_pass, high_pass]
    b, a = butter(buttord, window, btype='bandpass', fs=samplerate)
    return filtfilt(b, a, broadband_data, axis=0)

def filter_kinematics(kinematic_data, samplerate, low_cut=15, buttord=4):
    '''
    Low-pass filter kinematics data to below 15 Hz by default. 
    Filter parameters taken from Bradberry, et al., 2009 
    https://doi.org/10.1016/j.neuroimage.2009.06.023 

    Args:
        kinematic_data (nt, ...): raw hand data, e.g.
        samplerate (float): sampling rate of the raw data
        low_cut (float, optional): cutoff frequency for low-pass filter. Defaults to 15 Hz
        buttord (int, optional): order for butterworth low-pass filter. Defaults to 4.

    Returns:
        tuple: tuple containing:
        | **filtere_data (nt, ...):** filtered kinematics data
        | **samplerate (float):** sampling rate of the kinematics data

    Examples:

        .. code-block:: python

            fs = 100
            x_single, t = utils.generate_test_signal(T=5, fs, 1, 5)
            x_noise, t = utils.generate_test_signal(T=5, fs=fs, freq=[1,3,30], a=[5, 2, 0.5], noise=0.2)
            x_filt, _ = precondition.filter_kinematics(x, fs, low_cut=15, buttord=4)
            fig, ax = plt.subplot_mosaic([['A', 'B'],['C', 'C']])
        
            ax['A'].plot(t, x_noise, label='Noisy signal')
            ax['A'].plot(t, x_filt, label='Filtered signal')
            ax['A'].set_xlabel('time (seconds)')
            
            x_filt_simple = filt_fun(x_single)
            ax['B'].plot(t, x_single, label=f'{freq[0]} Hz signal')
            ax['B'].plot(t, x_filt_simple, label='Filtered signal')
            ax['B'].set_xlabel('time (seconds)')

            f_noise, psd_noise = analysis.get_psd_welch(x_noise, fs)
            f_filt, psd_filt = analysis.get_psd_welch(x_filt, fs)
            ax['C'].semilogy(f_noise, psd_noise, label='Noisy signal')
            ax['C'].semilogy(f_filt, psd_filt, label='Filtered signal')
            ax['C'].set_xlabel('frequency (Hz)')
            ax['C'].set_ylabel('PSD')

        .. image:: _images/filter_kinematics.png
    '''
    b, a = butter(buttord, low_cut, btype='lowpass', fs=samplerate)
    filtered_data = filtfilt(b, a, kinematic_data, axis=0)
    return filtered_data, samplerate
